# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DatasetBuilder base class."""

from __future__ import annotations

import abc
from collections.abc import Sequence
import dataclasses
import functools
import inspect
import json
import os
import sys
import typing
from typing import Any, ClassVar, Dict, Iterable, Iterator, List, Optional, Tuple, Type, Union

from absl import logging
from etils import epath
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import dataset_metadata
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import download
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import logging as tfds_logging
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import reader as reader_lib
from tensorflow_datasets.core import registered
from tensorflow_datasets.core import split_builder as split_builder_lib
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import tf_compat
from tensorflow_datasets.core import units
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.data_sources import array_record
from tensorflow_datasets.core.proto import dataset_info_pb2
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils import gcs_utils
from tensorflow_datasets.core.utils import read_config as read_config_lib
from tensorflow_datasets.core.utils import tree_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import termcolor


if typing.TYPE_CHECKING:
  import apache_beam as beam
  from apache_beam.runners import runner

ListOrTreeOrElem = type_utils.ListOrTreeOrElem
Tree = type_utils.Tree
TreeDict = type_utils.TreeDict
VersionOrStr = Union[utils.Version, str]

FORCE_REDOWNLOAD = download.GenerateMode.FORCE_REDOWNLOAD
REUSE_CACHE_IF_EXISTS = download.GenerateMode.REUSE_CACHE_IF_EXISTS
REUSE_DATASET_IF_EXISTS = download.GenerateMode.REUSE_DATASET_IF_EXISTS
UPDATE_DATASET_INFO = download.GenerateMode.UPDATE_DATASET_INFO

GCS_HOSTED_MSG = """\
Dataset %s is hosted on GCS. It will automatically be downloaded to your
local data directory. If you'd instead prefer to read directly from our public
GCS bucket (recommended if you're running on GCP), you can instead pass
`try_gcs=True` to `tfds.load` or set `data_dir=gs://tfds-data/datasets`.
"""


@dataclasses.dataclass(eq=False)
class BuilderConfig:
  """Base class for `DatasetBuilder` data configuration.

  DatasetBuilder subclasses with data configuration options should subclass
  `BuilderConfig` and add their own properties.

  Attributes:
    name: The name of the config.
    version: The version of the config.
    release_notes: A dictionary associating versions to changes.
    supported_versions: A list of versions which this Builder Config supports.
    description: a human description of the config.
    tags: [Experimental] a list of freeform tags applying to the config. This is
      not used by TFDS, but can be retrieved later from a ConfigBuilder
      instance.
  """

  # TODO(py3.10): Should update dataclass to be:
  # * Frozen (https://bugs.python.org/issue32953)
  # * Kwargs-only (https://bugs.python.org/issue33129)

  name: str
  version: Optional[VersionOrStr] = None
  release_notes: Optional[Dict[str, str]] = None
  supported_versions: List[VersionOrStr] = dataclasses.field(
      default_factory=list
  )
  description: Optional[str] = None
  tags: List[str] = dataclasses.field(default_factory=list)

  @classmethod
  def from_dataset_info(
      cls,
      info_proto: dataset_info_pb2.DatasetInfo,
  ) -> Optional["BuilderConfig"]:
    """Instantiates a BuilderConfig from the given proto.

    Args:
      info_proto: DatasetInfo proto which documents the requested dataset
        config, including its name, version, and features.

    Returns:
      A BuilderConfig for the requested config.
    """
    if not info_proto.config_name:
      return None
    return BuilderConfig(
        name=info_proto.config_name,
        description=info_proto.config_description,
        version=info_proto.version,
        release_notes=info_proto.release_notes or {},
        tags=info_proto.config_tags or [],
    )


def _get_builder_datadir_path(builder_cls: Type[Any]) -> epath.Path:
  """Returns the path to ConfigBuilder data dir.

  Args:
    builder_cls: a Builder class.

  Returns:
    The path to directory where the config data files of `builders_cls` can be
    read from. e.g. "/usr/lib/[...]/tensorflow_datasets/datasets/foo".
  """
  pkg_names = builder_cls.__module__.split(".")
  # -1 to remove the xxx.py python file.
  return epath.resource_path(pkg_names[0]).joinpath(*pkg_names[1:-1])


class DatasetBuilder(registered.RegisteredDataset):
  """Abstract base class for all datasets.

  `DatasetBuilder` has 3 key methods:

    * `DatasetBuilder.info`: documents the dataset, including feature
      names, types, and shapes, version, splits, citation, etc.
    * `DatasetBuilder.download_and_prepare`: downloads the source data
      and writes it to disk.
    * `DatasetBuilder.as_dataset`: builds an input pipeline using
      `tf.data.Dataset`s.

  **Configuration**: Some `DatasetBuilder`s expose multiple variants of the
  dataset by defining a `tfds.core.BuilderConfig` subclass and accepting a
  config object (or name) on construction. Configurable datasets expose a
  pre-defined set of configurations in `DatasetBuilder.builder_configs`.

  Typical `DatasetBuilder` usage:

  ```python
  mnist_builder = tfds.builder("mnist")
  mnist_info = mnist_builder.info
  mnist_builder.download_and_prepare()
  datasets = mnist_builder.as_dataset()

  train_dataset, test_dataset = datasets["train"], datasets["test"]
  assert isinstance(train_dataset, tf.data.Dataset)

  # And then the rest of your input pipeline
  train_dataset = train_dataset.repeat().shuffle(1024).batch(128)
  train_dataset = train_dataset.prefetch(2)
  features = tf.compat.v1.data.make_one_shot_iterator(train_dataset).get_next()
  image, label = features['image'], features['label']
  ```
  """

  # Semantic version of the dataset (ex: tfds.core.Version('1.2.0'))
  VERSION = None

  # Release notes
  # Metadata only used for documentation. Should be a dict[version,description]
  # Multi-lines are automatically dedent
  RELEASE_NOTES: ClassVar[Dict[str, str]] = {}

  # List dataset versions which can be loaded using current code.
  # Data can only be prepared with canonical VERSION or above.
  SUPPORTED_VERSIONS = []

  # Named configurations that modify the data generated by download_and_prepare.
  BUILDER_CONFIGS = []

  # Name of the builder config that should be used in case the user doesn't
  # specify a config when loading a dataset. If None, then the first config in
  # `BUILDER_CONFIGS` is used.
  DEFAULT_BUILDER_CONFIG_NAME: Optional[str] = None

  # Must be set for datasets that use 'manual_dir' functionality - the ones
  # that require users to do additional steps to download the data
  # (this is usually due to some external regulations / rules).
  #
  # This field should contain a string with user instructions, including
  # the list of files that should be present. It will be
  # displayed in the dataset documentation.
  MANUAL_DOWNLOAD_INSTRUCTIONS = None

  # Optional max number of simultaneous downloads. Setting this value will
  # override download config settings if necessary.
  MAX_SIMULTANEOUS_DOWNLOADS: Optional[int] = None

  # If not set, pkg_dir_path is inferred. However, if user of class knows better
  # then this can be set directly before init, to avoid heuristic inferences.
  # Example: `imported_builder_cls` function in `registered.py` module sets it.
  pkg_dir_path: Optional[epath.Path] = None

  @classmethod
  def _get_pkg_dir_path(cls) -> epath.Path:
    """Returns class pkg_dir_path, infer it first if not set."""
    # We use cls.__dict__.get to check whether `pkg_dir_path` attribute has been
    # set on the class, and not on a parent class. If we were accessing the
    # attribute directly, a dataset Builder inheriting from another would read
    # its metadata from its parent directory (which would be wrong).
    if cls.pkg_dir_path is None:
      cls.pkg_dir_path = _get_builder_datadir_path(cls)
    return cls.pkg_dir_path

  @classmethod
  def get_metadata(cls) -> dataset_metadata.DatasetMetadata:
    """Returns metadata (README, CITATIONS, ...) specified in config files.

    The config files are read from the same package where the DatasetBuilder has
    been defined, so those metadata might be wrong for legacy builders.
    """
    return dataset_metadata.load(cls._get_pkg_dir_path())

  @tfds_logging.builder_init()
  def __init__(
      self,
      *,
      data_dir: Optional[epath.PathLike] = None,
      config: Union[None, str, BuilderConfig] = None,
      version: Union[None, str, utils.Version] = None,
  ):
    """Constructs a DatasetBuilder.

    Callers must pass arguments as keyword arguments.

    Args:
      data_dir: directory to read/write data. Defaults to the value of the
        environment variable TFDS_DATA_DIR, if set, otherwise falls back to
        datasets are stored.
      config: `tfds.core.BuilderConfig` or `str` name, optional configuration
        for the dataset that affects the data generated on disk. Different
        `builder_config`s will have their own subdirectories and versions.
      version: Optional version at which to load the dataset. An error is raised
        if specified version cannot be satisfied. Eg: '1.2.3', '1.2.*'. The
        special value "experimental_latest" will use the highest version, even
        if not default. This is not recommended unless you know what you are
        doing, as the version could be broken.
    """
    if data_dir:
      data_dir = os.fspath(data_dir)  # Pathlib -> str
    # For pickling:
    self._original_state = dict(
        data_dir=data_dir, config=config, version=version
    )
    # To do the work:
    self._builder_config = self._create_builder_config(config)
    # Extract code version (VERSION or config)
    self._version = self._pick_version(version)
    # Compute the base directory (for download) and dataset/version directory.
    self._data_dir_root, self._data_dir = self._build_data_dir(data_dir)
    if self.data_path.exists():
      self.info.read_from_directory(self._data_dir)
    else:  # Use the code version (do not restore data)
      self.info.initialize_from_bucket()

  @utils.classproperty
  @classmethod
  @utils.memoize()
  def code_path(cls) -> Optional[epath.Path]:
    """Returns the path to the file where the Dataset class is located.

    Note: As the code can be run inside zip file. The returned value is
    a `Path` by default. Use `tfds.core.utils.to_write_path()` to cast
    the path into `Path`.

    Returns:
      path: pathlib.Path like abstraction
    """
    modules = cls.__module__.split(".")
    if len(modules) >= 2:  # Filter `__main__`, `python my_dataset.py`,...
      # If the dataset can be loaded from a module, use this to support zipapp.
      # Note: `utils.resource_path` will return either `zipfile.Path` (for
      # zipapp) or `pathlib.Path`.
      try:
        path = epath.resource_path(modules[0])
      except TypeError:  # Module is not a package
        pass
      else:
        # For dynamically added modules, `importlib.resources` returns
        # `pathlib.Path('.')` rather than the real path, so filter those by
        # checking for `parts`.
        # Check for `zipfile.Path` (`ResourcePath`) as it does not have `.parts`
        if isinstance(path, epath.resource_utils.ResourcePath) or path.parts:
          modules[-1] += ".py"
          return path.joinpath(*modules[1:])
    # Otherwise, fallback to `pathlib.Path`. For non-zipapp, it should be
    # equivalent to the above return.
    try:
      filepath = inspect.getfile(cls)
    except TypeError:
      # Could happen when the class is defined in Colab.
      return None
    else:
      return epath.Path(filepath)

  def __getstate__(self):
    return self._original_state

  def __setstate__(self, state):
    self.__init__(**state)

  @functools.cached_property
  def canonical_version(self) -> utils.Version:
    return canonical_version_for_config(self, self._builder_config)

  @functools.cached_property
  def supported_versions(self):
    if self._builder_config and self._builder_config.supported_versions:
      return self._builder_config.supported_versions
    else:
      return self.SUPPORTED_VERSIONS

  @functools.cached_property
  def versions(self) -> List[utils.Version]:
    """Versions (canonical + availables), in preference order."""
    return [
        utils.Version(v) if isinstance(v, str) else v
        for v in [self.canonical_version] + self.supported_versions
    ]

  def _pick_version(
      self, requested_version: Union[str, utils.Version]
  ) -> utils.Version:
    """Returns utils.Version instance, or raise AssertionError."""
    # Validate that `canonical_version` is correctly defined
    assert self.canonical_version
    # If requested_version is of type utils.Version, convert to its string
    # representation. This is necessary to properly execute the equality check
    # with "experimental_latest".
    if isinstance(requested_version, utils.Version):
      requested_version = str(requested_version)
    if requested_version == "experimental_latest":
      return max(self.versions)
    for version in self.versions:
      if requested_version is None or version.match(requested_version):
        return version
    available_versions = [str(v) for v in self.versions]
    msg = "Dataset {} cannot be loaded at version {}, only: {}.".format(
        self.name, requested_version, ", ".join(available_versions)
    )
    raise AssertionError(msg)

  @property
  def version(self) -> utils.Version:
    return self._version

  @property
  def release_notes(self) -> Dict[str, str]:
    if self.builder_config and self.builder_config.release_notes:
      return self.builder_config.release_notes
    else:
      return self.RELEASE_NOTES

  @property
  def data_dir_root(self) -> epath.Path:
    """Returns the root directory where all TFDS datasets are stored.

    Note that this is different from `data_dir`, which includes the dataset
    name, config, and version. For example, if `data_dir` is
    `/data/tfds/my_dataset/my_config/1.2.3`, then `data_dir_root` is
    `/data/tfds`.

    Returns:
      the root directory where all TFDS datasets are stored.
    """
    return epath.Path(self._data_dir_root)

  @property
  def data_dir(self) -> str:
    """Returns the directory where this version + config is stored.

    Note that this is different from `data_dir_root`. For example, if
    `data_dir_root` is `/data/tfds`, then `data_dir` would be
    `/data/tfds/my_dataset/my_config/1.2.3`.

    Returns:
      the directory where this version + config is stored.
    """
    return self._data_dir

  @property
  def data_path(self) -> epath.Path:
    """Returns the path where this version + config is stored."""
    # Instead, should make `_data_dir` be Path everywhere
    return epath.Path(self.data_dir)

  @utils.classproperty
  @classmethod
  def _checksums_path(cls) -> Optional[epath.Path]:
    """Returns the checksums path."""
    # Used:
    # * To load the checksums (in url_infos)
    # * To save the checksums (in DownloadManager)
    if not cls.code_path:
      return None
    new_path = cls.code_path.parent / "checksums.tsv"
    # Checksums of legacy datasets are located in a separate dir.
    legacy_path = utils.tfds_path() / "url_checksums" / f"{cls.name}.txt"
    if (
        # zipfile.Path does not have `.parts`. Additionally, `os.fspath`
        # will extract the file, so use `str`.
        "tensorflow_datasets" in str(new_path)
        and legacy_path.exists()
        and not new_path.exists()
    ):
      return legacy_path
    else:
      return new_path

  @utils.classproperty
  @classmethod
  @functools.lru_cache(maxsize=None)
  def url_infos(cls) -> Optional[Dict[str, download.checksums.UrlInfo]]:
    """Load `UrlInfo` from the given path."""
    # Note: If the dataset is downloaded with `record_checksums=True`, urls
    # might be updated but `url_infos` won't as it is memoized.

    # Search for the url_info file.
    checksums_path = cls._checksums_path
    # If url_info file is found, load the urls
    if checksums_path and checksums_path.exists():
      return download.checksums.load_url_infos(checksums_path)
    else:
      return None

  @property
  @tfds_logging.builder_info()
  # Warning: This unbounded cache is required for correctness. See b/238762111.
  @functools.lru_cache(maxsize=None)
  def info(self) -> dataset_info.DatasetInfo:
    """`tfds.core.DatasetInfo` for this builder."""
    # Ensure .info hasn't been called before versioning is set-up
    # Otherwise, backward compatibility cannot be guaranteed as some code will
    # depend on the code version instead of the restored data version
    if not getattr(self, "_version", None):
      # Message for developers creating new dataset. Will trigger if they are
      # using .info in the constructor before calling super().__init__
      raise AssertionError(
          "Info should not been called before version has been defined. "
          "Otherwise, the created .info may not match the info version from "
          "the restored dataset."
      )
    info = self._info()
    if not isinstance(info, dataset_info.DatasetInfo):
      raise TypeError(
          "DatasetBuilder._info should returns `tfds.core.DatasetInfo`, not "
          f" {type(info)}."
      )
    return info

  @utils.classproperty
  @classmethod
  def default_builder_config(cls) -> Optional[BuilderConfig]:
    return _get_default_config(
        builder_configs=cls.BUILDER_CONFIGS,
        default_config_name=cls.DEFAULT_BUILDER_CONFIG_NAME,
    )

  def get_default_builder_config(self) -> Optional[BuilderConfig]:
    """Returns the default builder config if there is one.

    Note that for dataset builders that cannot use the `cls.BUILDER_CONFIGS`, we
    need a method that uses the instance to get `BUILDER_CONFIGS` and
    `DEFAULT_BUILDER_CONFIG_NAME`.

    Returns:
      the default builder config if there is one
    """
    return _get_default_config(
        builder_configs=self.BUILDER_CONFIGS,
        default_config_name=self.DEFAULT_BUILDER_CONFIG_NAME,
    )

  def get_reference(
      self,
      namespace: Optional[str] = None,
  ) -> naming.DatasetReference:
    """Returns a reference to the dataset produced by this dataset builder.

    Includes the config if specified, the version, and the data_dir that should
    contain this dataset.

    Arguments:
      namespace: if this dataset is a community dataset, and therefore has a
        namespace, then the namespace must be provided such that it can be set
        in the reference. Note that a dataset builder is not aware that it is
        part of a namespace.

    Returns:
      a reference to this instantiated builder.
    """
    if self.builder_config:
      config = self.builder_config.name
    else:
      config = None
    return naming.DatasetReference(
        dataset_name=self.name,
        namespace=namespace,
        config=config,
        version=self.version,
        data_dir=self.data_dir_root,
    )

  def is_prepared(self) -> bool:
    """Returns whether this dataset is already downloaded and prepared."""
    return self.data_path.exists()

  @tfds_logging.download_and_prepare()
  def download_and_prepare(
      self,
      *,
      download_dir: Optional[epath.PathLike] = None,
      download_config: Optional[download.DownloadConfig] = None,
      file_format: Optional[Union[str, file_adapters.FileFormat]] = None,
  ) -> None:
    """Downloads and prepares dataset for reading.

    Args:
      download_dir: directory where downloaded files are stored. Defaults to
        "~/tensorflow-datasets/downloads".
      download_config: `tfds.download.DownloadConfig`, further configuration for
        downloading and preparing dataset.
      file_format: optional `str` or `file_adapters.FileFormat`, format of the
        record files in which the dataset will be written.

    Raises:
      IOError: if there is not enough disk space available.
      RuntimeError: when the config cannot be found.
    """

    download_config = download_config or download.DownloadConfig()
    data_path = self.data_path
    data_exists = data_path.exists()

    if download_config.download_mode == UPDATE_DATASET_INFO:
      self._update_dataset_info()
      return

    if data_exists and download_config.download_mode == REUSE_DATASET_IF_EXISTS:
      logging.info("Reusing dataset %s (%s)", self.name, self.data_dir)
      return
    elif data_exists and download_config.download_mode == REUSE_CACHE_IF_EXISTS:
      logging.info(
          "Deleting pre-existing dataset %s (%s)", self.name, self.data_dir
      )
      data_path.rmtree()  # Delete pre-existing data.
      data_exists = data_path.exists()

    if self.version.tfds_version_to_prepare:
      available_to_prepare = ", ".join(
          str(v) for v in self.versions if not v.tfds_version_to_prepare
      )
      raise AssertionError(
          "The version of the dataset you are trying to use ({}:{}) can only "
          "be generated using TFDS code synced @ {} or earlier. Either sync to "
          "that version of TFDS to first prepare the data or use another "
          "version of the dataset (available for `download_and_prepare`: "
          "{}).".format(
              self.name,
              self.version,
              self.version.tfds_version_to_prepare,
              available_to_prepare,
          )
      )

    # Only `cls.VERSION` or `experimental_latest` versions can be generated.
    # Otherwise, users may accidentally generate an old version using the
    # code from newer versions.
    installable_versions = {
        str(v) for v in (self.canonical_version, max(self.versions))
    }
    if str(self.version) not in installable_versions:
      msg = (
          "The version of the dataset you are trying to use ({}) is too "
          "old for this version of TFDS so cannot be generated."
      ).format(self.info.full_name)
      if self.version.tfds_version_to_prepare:
        msg += (
            "{} can only be generated using TFDS code synced @ {} or earlier "
            "Either sync to that version of TFDS to first prepare the data or "
            "use another version of the dataset. "
        ).format(self.version, self.version.tfds_version_to_prepare)
      else:
        msg += (
            "Either sync to a previous version of TFDS to first prepare the "
            "data or use another version of the dataset. "
        )
      msg += "Available for `download_and_prepare`: {}".format(
          list(sorted(installable_versions))
      )
      raise ValueError(msg)

    # Currently it's not possible to overwrite the data because it would
    # conflict with versioning: If the last version has already been generated,
    # it will always be reloaded and data_dir will be set at construction.
    if data_exists:
      raise ValueError(
          "Trying to overwrite an existing dataset {} at {}. A dataset with "
          "the same version {} already exists. If the dataset has changed, "
          "please update the version number.".format(
              self.name, self.data_dir, self.version
          )
      )

    logging.info("Generating dataset %s (%s)", self.name, self.data_dir)
    if not utils.has_sufficient_disk_space(
        self.info.dataset_size + self.info.download_size,
        directory=os.fspath(self.data_dir_root),
    ):
      raise IOError(
          "Not enough disk space. Needed: {} (download: {}, generated: {})"
          .format(
              self.info.dataset_size + self.info.download_size,
              self.info.download_size,
              self.info.dataset_size,
          )
      )
    self._log_download_bytes()

    dl_manager = self._make_download_manager(
        download_dir=download_dir,
        download_config=download_config,
    )

    # Maybe save the `builder_cls` metadata common to all builder configs.
    if self.BUILDER_CONFIGS:
      default_builder_config = self.get_default_builder_config()
      if default_builder_config is None:
        raise RuntimeError(
            "Could not find default builder config while there "
            "are builder configs!"
        )
      _save_default_config_name(
          # `data_dir/ds_name/config/version/` -> `data_dir/ds_name/`
          common_dir=self.data_path.parent.parent,
          default_config_name=default_builder_config.name,
      )

    # If the file format was specified, set it in the info such that it is used
    # to generate the files.
    if file_format:
      self.info.set_file_format(file_format, override=True)

    # Create a tmp dir and rename to self.data_dir on successful exit.
    with utils.incomplete_dir(self.data_dir) as tmp_data_dir:
      # Temporarily assign _data_dir to tmp_data_dir to avoid having to forward
      # it to every sub function.
      with utils.temporary_assignment(self, "_data_dir", tmp_data_dir):
        if download_config.try_download_gcs and gcs_utils.is_dataset_on_gcs(
            self.info.full_name
        ):
          logging.info(GCS_HOSTED_MSG, self.name)
          gcs_utils.download_gcs_dataset(
              dataset_name=self.info.full_name, local_dataset_dir=self.data_dir
          )
          self.info.read_from_directory(self.data_dir)
        else:
          self._download_and_prepare(
              dl_manager=dl_manager,
              download_config=download_config,
          )

          # NOTE: If modifying the lines below to put additional information in
          # DatasetInfo, you'll likely also want to update
          # DatasetInfo.read_from_directory to possibly restore these attributes
          # when reading from package data.
          self.info.download_size = dl_manager.downloaded_size
          # Write DatasetInfo to disk, even if we haven't computed statistics.
          self.info.write_to_directory(self.data_dir)
      # The generated DatasetInfo contains references to `tmp_data_dir`
      self.info.update_data_dir(self.data_dir)

    # Clean up incomplete files from preempted workers.
    deleted_incomplete_files = []
    for f in self.data_path.glob(f"*{constants.INCOMPLETE_SUFFIX}*"):
      if utils.is_incomplete_file(f):
        deleted_incomplete_files.append(os.fspath(f))
        f.unlink()
    if deleted_incomplete_files:
      logging.info(
          "Deleted %d incomplete files. A small selection: %s",
          len(deleted_incomplete_files),
          "\n".join(deleted_incomplete_files[:3]),
      )

    self._log_download_done()


  def _update_dataset_info(self) -> None:
    """Updates the `dataset_info.json` file in the dataset dir."""
    info_file = self.data_path / constants.DATASET_INFO_FILENAME
    if not info_file.exists():
      raise AssertionError(f"To update {info_file}, it must already exist.")
    new_info = self.info
    new_info.read_from_directory(self.data_path)
    new_info.write_to_directory(self.data_path, all_metadata=False)

  @tfds_logging.as_data_source()
  def as_data_source(
      self,
      split: Optional[Tree[splits_lib.SplitArg]] = None,
      *,
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]] = None,
  ) -> ListOrTreeOrElem[Sequence[Any]]:
    """Constructs an `ArrayRecordDataSource`.

    Args:
      split: Which split of the data to load (e.g. `'train'`, `'test'`,
        `['train', 'test']`, `'train[80%:]'`,...). See our [split API
        guide](https://www.tensorflow.org/datasets/splits). If `None`, will
        return all splits in a `Dict[Split, Sequence]`.
      decoders: Nested dict of `Decoder` objects which allow to customize the
        decoding. The structure should match the feature structure, but only
        customized feature keys need to be present. See [the
        guide](https://github.com/tensorflow/datasets/blob/master/docs/decode.md)
        for more info.

    Returns:
      `Sequence` if `split`,
      `dict<key: tfds.Split, value: Sequence>` otherwise.

    Raises:
      NotImplementedError if the data was not generated using ArrayRecords.
    """
    # By default, return all splits
    if split is None:
      split = {s: s for s in self.info.splits}

    # Create a dataset for each of the given splits
    def build_single_data_source(
        split: str,
    ) -> Sequence[Any]:
      return array_record.ArrayRecordDataSource(
          self.info,
          split=split,
          decoders=decoders,
      )

    all_ds = tree_utils.map_structure(build_single_data_source, split)
    return all_ds

  @tfds_logging.as_dataset()
  def as_dataset(
      self,
      split: Optional[Tree[splits_lib.SplitArg]] = None,
      *,
      batch_size: Optional[int] = None,
      shuffle_files: bool = False,
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]] = None,
      read_config: Optional[read_config_lib.ReadConfig] = None,
      as_supervised: bool = False,
  ):
    # pylint: disable=line-too-long
    """Constructs a `tf.data.Dataset`.

    Callers must pass arguments as keyword arguments.

    The output types vary depending on the parameters. Examples:

    ```python
    builder = tfds.builder('imdb_reviews')
    builder.download_and_prepare()

    # Default parameters: Returns the dict of tf.data.Dataset
    ds_all_dict = builder.as_dataset()
    assert isinstance(ds_all_dict, dict)
    print(ds_all_dict.keys())  # ==> ['test', 'train', 'unsupervised']

    assert isinstance(ds_all_dict['test'], tf.data.Dataset)
    # Each dataset (test, train, unsup.) consists of dictionaries
    # {'label': <tf.Tensor: .. dtype=int64, numpy=1>,
    #  'text': <tf.Tensor: .. dtype=string, numpy=b"I've watched the movie ..">}
    # {'label': <tf.Tensor: .. dtype=int64, numpy=1>,
    #  'text': <tf.Tensor: .. dtype=string, numpy=b'If you love Japanese ..'>}

    # With as_supervised: tf.data.Dataset only contains (feature, label) tuples
    ds_all_supervised = builder.as_dataset(as_supervised=True)
    assert isinstance(ds_all_supervised, dict)
    print(ds_all_supervised.keys())  # ==> ['test', 'train', 'unsupervised']

    assert isinstance(ds_all_supervised['test'], tf.data.Dataset)
    # Each dataset (test, train, unsup.) consists of tuples (text, label)
    # (<tf.Tensor: ... dtype=string, numpy=b"I've watched the movie ..">,
    #  <tf.Tensor: ... dtype=int64, numpy=1>)
    # (<tf.Tensor: ... dtype=string, numpy=b"If you love Japanese ..">,
    #  <tf.Tensor: ... dtype=int64, numpy=1>)

    # Same as above plus requesting a particular split
    ds_test_supervised = builder.as_dataset(as_supervised=True, split='test')
    assert isinstance(ds_test_supervised, tf.data.Dataset)
    # The dataset consists of tuples (text, label)
    # (<tf.Tensor: ... dtype=string, numpy=b"I've watched the movie ..">,
    #  <tf.Tensor: ... dtype=int64, numpy=1>)
    # (<tf.Tensor: ... dtype=string, numpy=b"If you love Japanese ..">,
    #  <tf.Tensor: ... dtype=int64, numpy=1>)
    ```

    Args:
      split: Which split of the data to load (e.g. `'train'`, `'test'`,
        `['train', 'test']`, `'train[80%:]'`,...). See our [split API
        guide](https://www.tensorflow.org/datasets/splits). If `None`, will
        return all splits in a `Dict[Split, tf.data.Dataset]`.
      batch_size: `int`, batch size. Note that variable-length features will be
        0-padded if `batch_size` is set. Users that want more custom behavior
        should use `batch_size=None` and use the `tf.data` API to construct a
        custom pipeline. If `batch_size == -1`, will return feature dictionaries
        of the whole dataset with `tf.Tensor`s instead of a `tf.data.Dataset`.
      shuffle_files: `bool`, whether to shuffle the input files. Defaults to
        `False`.
      decoders: Nested dict of `Decoder` objects which allow to customize the
        decoding. The structure should match the feature structure, but only
        customized feature keys need to be present. See [the
        guide](https://github.com/tensorflow/datasets/blob/master/docs/decode.md)
        for more info.
      read_config: `tfds.ReadConfig`, Additional options to configure the input
        pipeline (e.g. seed, num parallel reads,...).
      as_supervised: `bool`, if `True`, the returned `tf.data.Dataset` will have
        a 2-tuple structure `(input, label)` according to
        `builder.info.supervised_keys`. If `False`, the default, the returned
        `tf.data.Dataset` will have a dictionary with all the features.

    Returns:
      `tf.data.Dataset`, or if `split=None`, `dict<key: tfds.Split, value:
      tf.data.Dataset>`.

      If `batch_size` is -1, will return feature dictionaries containing
      the entire dataset in `tf.Tensor`s instead of a `tf.data.Dataset`.
    """
    # pylint: enable=line-too-long
    if not self.data_path.exists():
      raise AssertionError(
          "Dataset %s: could not find data in %s. Please make sure to call "
          "dataset_builder.download_and_prepare(), or pass download=True to "
          "tfds.load() before trying to access the tf.data.Dataset object."
          % (self.name, self.data_dir_root)
      )

    # By default, return all splits
    if split is None:
      split = {s: s for s in self.info.splits}

    read_config = read_config or read_config_lib.ReadConfig()

    # Create a dataset for each of the given splits
    build_single_dataset = functools.partial(
        self._build_single_dataset,
        shuffle_files=shuffle_files,
        batch_size=batch_size,
        decoders=decoders,
        read_config=read_config,
        as_supervised=as_supervised,
    )
    all_ds = tree_utils.map_structure(build_single_dataset, split)
    return all_ds

  def _build_single_dataset(
      self,
      split: splits_lib.Split,
      batch_size: Optional[int],
      shuffle_files: bool,
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]],
      read_config: read_config_lib.ReadConfig,
      as_supervised: bool,
  ) -> tf.data.Dataset:
    """as_dataset for a single split."""
    wants_full_dataset = batch_size == -1
    if wants_full_dataset:
      batch_size = self.info.splits.total_num_examples or sys.maxsize

    # Build base dataset
    ds = self._as_dataset(
        split=split,
        shuffle_files=shuffle_files,
        decoders=decoders,
        read_config=read_config,
    )
    # Auto-cache small datasets which are small enough to fit in memory.
    if self._should_cache_ds(
        split=split, shuffle_files=shuffle_files, read_config=read_config
    ):
      ds = ds.cache()

    if batch_size:
      # Use padded_batch so that features with unknown shape are supported.
      ds = ds.padded_batch(batch_size, tf.compat.v1.data.get_output_shapes(ds))

    if as_supervised:
      if not self.info.supervised_keys:
        raise ValueError(
            f"as_supervised=True but {self.name} does not support a supervised "
            "structure."
        )

      def lookup_nest(features: Dict[str, Any]) -> Tuple[Any, ...]:
        """Converts `features` to the structure described by `supervised_keys`.

        Note that there is currently no way to access features in nested
        feature dictionaries.

        Args:
          features: dictionary of features

        Returns:
          A tuple with elements structured according to `supervised_keys`
        """
        return tree_utils.map_structure(
            lambda key: features[key], self.info.supervised_keys
        )

      ds = ds.map(lookup_nest)

    # Add prefetch by default
    if not read_config.skip_prefetch:
      ds = ds.prefetch(tf.data.experimental.AUTOTUNE)

    if wants_full_dataset:
      return tf_compat.get_single_element(ds)
    return ds

  def _should_cache_ds(self, split, shuffle_files, read_config) -> bool:
    """Returns True if TFDS should auto-cache the dataset."""
    # The user can explicitly opt-out from auto-caching
    if not read_config.try_autocache:
      return False

    # Skip datasets with unknown size.
    # Even by using heuristic with `download_size` and
    # `MANUAL_DOWNLOAD_INSTRUCTIONS`, it wouldn't catch datasets which hardcode
    # the non-processed data-dir, nor DatasetBuilder not based on tf-record.
    if not self.info.dataset_size:
      return False

    # Do not cache big datasets
    # Instead of using the global size, we could infer the requested bytes:
    # `self.info.splits[split].num_bytes`
    # The info is available for full splits, and could be approximated
    # for subsplits `train[:50%]`.
    # However if the user is creating multiple small splits from a big
    # dataset, those could adds up and fill up the entire RAM.
    # 250 MiB is arbitrary picked. For comparison, Cifar10 is about 150 MiB.
    if self.info.dataset_size > 250 * units.MiB:
      return False

    # We do not want to cache data which has more than one shards when
    # shuffling is enabled, as this would effectively disable shuffling.
    # An exception is for single shard (as shuffling is a no-op).
    # Another exception is if reshuffle is disabled (shuffling already cached)
    num_shards = len(self.info.splits[split].file_instructions)
    if (
        shuffle_files
        and
        # Shuffling only matter when reshuffle is True or None (default)
        read_config.shuffle_reshuffle_each_iteration is not False
        and num_shards > 1  # pylint: disable=g-bool-id-comparison
    ):
      return False

    # If the dataset satisfy all the right conditions, activate autocaching.
    return True

  def _relative_data_dir(self, with_version: bool = True) -> str:
    """Relative path of this dataset in data_dir."""
    builder_data_dir = self.name
    builder_config = self._builder_config
    if builder_config:
      builder_data_dir = os.path.join(builder_data_dir, builder_config.name)
    if not with_version:
      return builder_data_dir

    version_data_dir = os.path.join(builder_data_dir, str(self._version))
    return version_data_dir

  def _build_data_dir(self, given_data_dir: Optional[str]):
    """Return the data directory for the current version.

    Args:
      given_data_dir: `Optional[str]`, root `data_dir` passed as `__init__`
        argument.

    Returns:
      data_dir_root: `str`, The root dir containing all datasets, downloads,...
      data_dir: `str`, The version data_dir
        (e.g. `<data_dir_root>/<ds_name>/<config>/<version>`)
    """
    builder_dir = self._relative_data_dir(with_version=False)
    version_dir = self._relative_data_dir(with_version=True)

    default_data_dir = file_utils.get_default_data_dir(
        given_data_dir=given_data_dir
    )
    all_data_dirs = file_utils.list_data_dirs(
        given_data_dir=given_data_dir, dataset=self.name
    )

    all_versions = set()
    requested_version_dirs = {}
    for data_dir_root in all_data_dirs:
      # List all existing versions
      full_builder_dir = os.path.join(data_dir_root, builder_dir)
      data_dir_versions = set(utils.version.list_all_versions(full_builder_dir))
      # Check for existence of the requested version
      if self.version in data_dir_versions:
        requested_version_dirs[data_dir_root] = os.path.join(
            data_dir_root, version_dir
        )
      all_versions.update(data_dir_versions)

    if len(requested_version_dirs) > 1:
      raise ValueError(
          "Dataset was found in more than one directory: {}. Please resolve "
          "the ambiguity by explicitly specifying `data_dir=`."
          "".format(requested_version_dirs.values())
      )
    elif len(requested_version_dirs) == 1:  # The dataset is found once
      return next(iter(requested_version_dirs.items()))

    # No dataset found, use default directory
    data_dir = os.path.join(default_data_dir, version_dir)
    if all_versions:
      logging.warning(
          (
              "Found a different version of the requested dataset:\n"
              "%s\n"
              "Using %s instead."
          ),
          "\n".join(str(v) for v in sorted(all_versions)),
          data_dir,
      )
    return default_data_dir, data_dir

  def _log_download_done(self) -> None:
    msg = (
        f"Dataset {self.name} downloaded and prepared to {self.data_dir}. "
        "Subsequent calls will reuse this data."
    )
    termcolor.cprint(msg, attrs=["bold"])

  def _log_download_bytes(self) -> None:
    # Print is intentional: we want this to always go to stdout so user has
    # information needed to cancel download/preparation if needed.
    # This comes right before the progress bar.
    termcolor.cprint(
        (
            f"Downloading and preparing dataset {self.info.download_size} "
            f"(download: {self.info.download_size}, "
            f"generated: {self.info.dataset_size}, "
            f"total: {self.info.download_size + self.info.dataset_size}) "
            f"to {self.data_dir}..."
        ),
        attrs=["bold"],
    )

  def dataset_info_from_configs(self, **kwargs):
    """Returns the DatasetInfo using given kwargs and config files.

    Sub-class should call this and add information not present in config files
    using kwargs directly passed to tfds.core.DatasetInfo object.

    If information is present both in passed arguments and config files, config
    files will prevail.

    Args:
      **kwargs: kw args to pass to DatasetInfo directly.
    """
    metadata = self.get_metadata()
    if metadata.description:
      kwargs["description"] = metadata.description
    if metadata.citation:
      kwargs["citation"] = metadata.citation
    return dataset_info.DatasetInfo(builder=self, **kwargs)

  @abc.abstractmethod
  @utils.docs.doc_private
  def _info(self) -> dataset_info.DatasetInfo:
    """Returns the `tfds.core.DatasetInfo` object.

    This function is called once and the result is cached for all
    following calls.

    Returns:
      dataset_info: The dataset metadata.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _download_and_prepare(
      self,
      dl_manager: download.DownloadManager,
      download_config: Optional[download.DownloadConfig] = None,
  ) -> None:
    """Downloads and prepares dataset for reading.

    Internal implementation to overwrite when inheriting from DatasetBuilder.
    Called when `builder.download_and_prepare` is called.
    It should download all required data and generate
    the pre-processed datasets files.

    Args:
      dl_manager: `tfds.download.DownloadManager` used to download and cache
        data.
      download_config: `DownloadConfig`, Additional options.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _as_dataset(
      self,
      split: splits_lib.Split,
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]] = None,
      read_config: Optional[read_config_lib.ReadConfig] = None,
      shuffle_files: bool = False,
  ) -> tf.data.Dataset:
    """Constructs a `tf.data.Dataset`.

    Internal implementation to overwrite when inheriting from DatasetBuilder.
    Called when `builder.as_dataset` is called.
    It should read the pre-processed datasets files and generate
    the `tf.data.Dataset` object.

    Args:
      split: `tfds.Split` which subset of the data to read.
      decoders: Nested structure of `Decoder` object to customize the dataset
        decoding.
      read_config: `tfds.ReadConfig`
      shuffle_files: `bool`, whether to shuffle the input files. Optional,
        defaults to `False`.

    Returns:
      `tf.data.Dataset`
    """
    raise NotImplementedError

  def _make_download_manager(
      self,
      download_dir: epath.PathLike | None,
      download_config: download.DownloadConfig,
  ) -> download.DownloadManager:
    """Creates a new download manager object."""
    download_dir = epath.Path(download_dir or self.data_dir_root / "downloads")
    extract_dir = download_config.extract_dir or download_dir / "extracted"
    manual_dir = download_config.manual_dir or download_dir / "manual"

    if download_config.register_checksums:
      # Note: Error will be raised here if user try to record checksums
      # from a `zipapp`
      register_checksums_path = utils.to_write_path(self._checksums_path)
    else:
      register_checksums_path = None

    max_simultaneous_downloads = (
        download_config.override_max_simultaneous_downloads
        or self.MAX_SIMULTANEOUS_DOWNLOADS
    )
    if (
        max_simultaneous_downloads
        and self.MAX_SIMULTANEOUS_DOWNLOADS
        and self.MAX_SIMULTANEOUS_DOWNLOADS < max_simultaneous_downloads
    ):
      logging.warning(
          (
              "The dataset %r sets `MAX_SIMULTANEOUS_DOWNLOADS`=%s. The"
              " download config overrides it with value"
              " `override_max_simultaneous_downloads`=%s. Using the higher"
              " value might cause `ConnectionRefusedError`"
          ),
          self.name,
          self.MAX_SIMULTANEOUS_DOWNLOADS,
          download_config.override_max_simultaneous_downloads,
      )

    return download.DownloadManager(
        download_dir=download_dir,
        extract_dir=extract_dir,
        manual_dir=manual_dir,
        url_infos=self.url_infos,
        manual_dir_instructions=self.MANUAL_DOWNLOAD_INSTRUCTIONS,
        force_download=(download_config.download_mode == FORCE_REDOWNLOAD),
        force_extraction=(download_config.download_mode == FORCE_REDOWNLOAD),
        force_checksums_validation=download_config.force_checksums_validation,
        register_checksums=download_config.register_checksums,
        register_checksums_path=register_checksums_path,
        verify_ssl=download_config.verify_ssl,
        dataset_name=self.name,
        max_simultaneous_downloads=max_simultaneous_downloads,
    )

  @utils.docs.do_not_doc_in_subclasses
  @utils.classproperty
  @classmethod
  def builder_config_cls(cls) -> Optional[type[BuilderConfig]]:
    """Returns the builder config class."""
    if not cls.BUILDER_CONFIGS:
      return None

    all_cls = {type(b) for b in cls.BUILDER_CONFIGS}
    if len(all_cls) != 1:
      raise ValueError(
          f"Could not infer the config class for {cls}. Detected: {all_cls}"
      )

    (builder_cls,) = all_cls
    return builder_cls

  @property
  def builder_config(self) -> Optional[Any]:
    """`tfds.core.BuilderConfig` for this builder."""
    return self._builder_config

  def _create_builder_config(self, builder_config) -> Optional[BuilderConfig]:
    """Create and validate BuilderConfig object."""
    if builder_config is None:
      builder_config = self.get_default_builder_config()
      if builder_config is not None:
        logging.info(
            "No config specified, defaulting to config: %s/%s",
            self.name,
            builder_config.name,
        )
    if not builder_config:
      return None
    if isinstance(builder_config, str):
      name = builder_config
      builder_config = self.builder_configs.get(name)
      if builder_config is None:
        raise ValueError(
            "BuilderConfig %s not found. Available: %s"
            % (name, list(self.builder_configs.keys()))
        )
    name = builder_config.name
    if not name:
      raise ValueError("BuilderConfig must have a name, got %s" % name)
    is_custom = name not in self.builder_configs
    if is_custom:
      logging.warning("Using custom data configuration %s", name)
    else:
      if builder_config is not self.builder_configs[name]:
        raise ValueError(
            "Cannot name a custom BuilderConfig the same as an available "
            "BuilderConfig. Change the name. Available BuilderConfigs: %s"
            % (list(self.builder_configs.keys()))
        )
    return builder_config

  @utils.classproperty
  @classmethod
  @utils.memoize()
  def builder_configs(cls):
    """Pre-defined list of configurations for this builder class."""
    config_dict = {config.name: config for config in cls.BUILDER_CONFIGS}
    if len(config_dict) != len(cls.BUILDER_CONFIGS):
      names = [config.name for config in cls.BUILDER_CONFIGS]
      raise ValueError(
          "Names in BUILDER_CONFIGS must not be duplicated. Got %s" % names
      )
    return config_dict


class FileReaderBuilder(DatasetBuilder):
  """Base class for datasets reading files.

  Subclasses are:

   * `GeneratorBasedBuilder`: Can both generate and read generated dataset.
   * `ReadOnlyBuilder`: Can only read pre-generated datasets. A user can
     generate a dataset with `GeneratorBasedBuilder`, and read them with
     `ReadOnlyBuilder` without requiring the original generation code.
  """

  @tfds_logging.builder_init()
  def __init__(
      self,
      *,
      file_format: Union[None, str, file_adapters.FileFormat] = None,
      **kwargs: Any,
  ):
    """Initializes an instance of FileReaderBuilder.

    Callers must pass arguments as keyword arguments.

    Args:
      file_format: EXPERIMENTAL, may change at any time; Format of the record
        files in which dataset will be read/written to. If `None`, defaults to
        `tfrecord`.
      **kwargs: Arguments passed to `DatasetBuilder`.
    """
    super().__init__(**kwargs)
    self.info.set_file_format(file_format)

  @functools.cached_property
  def _example_specs(self):
    return self.info.features.get_serialized_info()

  def _as_dataset(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self,
      split: splits_lib.Split,
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]],
      read_config: read_config_lib.ReadConfig,
      shuffle_files: bool,
  ) -> tf.data.Dataset:
    # Partial decoding
    # TODO(epot): Should be moved inside `features.decode_example`
    if isinstance(decoders, decode.PartialDecoding):
      features = decoders.extract_features(self.info.features)
      example_specs = features.get_serialized_info()
      decoders = decoders.decoders
    # Full decoding (all features decoded)
    else:
      features = self.info.features
      example_specs = self._example_specs
      decoders = decoders  # pylint: disable=self-assigning-variable

    reader = reader_lib.Reader(
        self.data_dir,
        example_specs=example_specs,
        file_format=self.info.file_format,
    )
    decode_fn = functools.partial(features.decode_example, decoders=decoders)
    return reader.read(
        instructions=split,
        split_infos=self.info.splits.values(),
        decode_fn=decode_fn,
        read_config=read_config,
        shuffle_files=shuffle_files,
        disable_shuffling=self.info.disable_shuffling,
    )


class GeneratorBasedBuilder(FileReaderBuilder):
  """Base class for datasets with data generation based on file adapter.

  `GeneratorBasedBuilder` is a convenience class that abstracts away much
  of the data writing and reading of `DatasetBuilder`.

  It expects subclasses to overwrite `_split_generators` to return a dict of
  splits, generators. See the method docstrings for details.
  """

  @abc.abstractmethod
  @utils.docs.do_not_doc_in_subclasses
  @utils.docs.doc_private
  def _split_generators(
      self,
      dl_manager: download.DownloadManager,
  ) -> Dict[splits_lib.Split, split_builder_lib.SplitGenerator]:
    """Downloads the data and returns dataset splits with associated examples.

    Example:

    ```python
    def _split_generators(self, dl_manager):
      path = dl_manager.download_and_extract('http://dataset.org/my_data.zip')
      return {
          'train': self._generate_examples(path=path / 'train_imgs'),
          'test': self._generate_examples(path=path / 'test_imgs'),
      }
    ```

    * If the original dataset do not have predefined `train`, `test`,... splits,
      this function should only returns a single `train` split here. Users can
      use the [subsplit API](https://www.tensorflow.org/datasets/splits) to
      create subsplits (e.g.
      `tfds.load(..., split=['train[:75%]', 'train[75%:]'])`).
    * `tfds.download.DownloadManager` caches downloads, so calling `download`
      on the same url multiple times only download it once.
    * A good practice is to download all data in this function, and have all the
      computation inside `_generate_examples`.
    * Splits are generated in the order defined here. `builder.info.splits` keep
      the same order.
    * This function can have an extra `pipeline` kwarg only if some
      beam preprocessing should be shared across splits. In this case,
      a dict of `beam.PCollection` should be returned.
      See `_generate_example` for details.

    Args:
      dl_manager: `tfds.download.DownloadManager` used to download/extract the
        data

    Returns:
      The dict of split name, generators. See `_generate_examples` for details
      about the generator format.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  @utils.docs.do_not_doc_in_subclasses
  @utils.docs.doc_private
  def _generate_examples(
      self, **kwargs: Any
  ) -> split_builder_lib.SplitGenerator:
    """Default function to generate examples for each split.

    The function should return a collection of `(key, examples)`. Examples
    will be encoded and written to disk. See `yields` section for details.

    The function can return/yield:

    * A python generator:

    ```python
    def _generate_examples(self, path):
      for filepath in path.iterdir():
        yield filepath.name, {'image': ..., 'label': ...}
    ```

    * A `beam.PTransform` of (input_types: [] -> output_types: `KeyExample`):
    For big datasets and distributed generation. See our Apache Beam
    [datasets guide](https://www.tensorflow.org/datasets/beam_datasets)
    for more info.

    ```python
    def _generate_examples(self, path):
      return (
          beam.Create(path.iterdir())
          | beam.Map(lambda filepath: filepath.name, {'image': ..., ...})
      )
    ```

    * A `beam.PCollection`: This should only be used if you need to share some
    distributed processing across splits. In this case, you can use the
    following pattern:

    ```python
    def _split_generators(self, dl_manager, pipeline):
      ...
      # Distributed processing shared across splits
      pipeline |= beam.Create(path.iterdir())
      pipeline |= 'SharedPreprocessing' >> beam.Map(_common_processing)
      ...
      # Wrap the pipeline inside a ptransform_fn to add `'label' >> ` and avoid
      # duplicated PTransform nodes names.
      generate_examples = beam.ptransform_fn(self._generate_examples)
      return {
          'train': pipeline | 'train' >> generate_examples(is_train=True)
          'test': pipeline | 'test' >> generate_examples(is_train=False)
      }

    def _generate_examples(self, pipeline, is_train: bool):
      return pipeline | beam.Map(_split_specific_processing, is_train=is_train)
    ```

    Note: Each split should uses a different tag name (e.g.
    `'train' >> generate_examples(path)`). Otherwise Beam will raise
    duplicated name error.

    Args:
      **kwargs: Arguments from the `_split_generators`

    Yields:
      key: `str` or `int`, a unique deterministic example identification key.
        * Unique: An error will be raised if two examples are yield with the
          same key.
        * Deterministic: When generating the dataset twice, the same example
          should have the same key.
        * Comparable: If shuffling is disabled the key will be used to sort the
        dataset.
        Good keys can be the image id, or line number if examples are extracted
        from a text file.
        The example will be sorted by `hash(key)` if shuffling is enabled, and
        otherwise by `key`.
        Generating the dataset multiple times will keep examples in the
        same order.
      example: `dict<str feature_name, feature_value>`, a feature dictionary
        ready to be encoded and written to disk. The example will be
        encoded with `self.info.features.encode_example({...})`.
    """
    raise NotImplementedError()

  def _process_pipeline_result(
      self, pipeline_result: "runner.PipelineResult"
  ) -> None:
    """Processes the result of the beam pipeline if we used one.

    This can be used to (e.g) write beam counters to a file.
    Args:
      pipeline_result: PipelineResult returned by beam.Pipeline.run().
    """
    return

  def _download_and_prepare(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self,
      dl_manager: download.DownloadManager,
      download_config: download.DownloadConfig,
  ) -> None:
    """Generate all splits and returns the computed split infos."""
    split_builder = split_builder_lib.SplitBuilder(
        split_dict=self.info.splits,
        features=self.info.features,
        dataset_size=self.info.dataset_size,
        max_examples_per_split=download_config.max_examples_per_split,
        beam_options=download_config.beam_options,
        beam_runner=download_config.beam_runner,
        file_format=self.info.file_format,
        shard_config=download_config.get_shard_config(),
    )
    # Wrap the generation inside a context manager.
    # If `beam` is used during generation (when a pipeline gets created),
    # the context manager is equivalent to `with beam.Pipeline()`.
    # Otherwise, this is a no-op.
    # By auto-detecting Beam, the user only has to change `_generate_examples`
    # to go from non-beam to beam dataset:
    # https://www.tensorflow.org/datasets/beam_datasets#instructions
    with split_builder.maybe_beam_pipeline() as maybe_pipeline_proxy:
      # If the signature has a `pipeline` kwargs, create the pipeline now and
      # forward it to `self._split_generators`
      # We add this magic because the pipeline kwargs is only used by c4 and
      # we do not want to make the API more verbose for a single advanced case.
      # See also the documentation at the end here:
      # https://www.tensorflow.org/datasets/api_docs/python/tfds/core/GeneratorBasedBuilder?version=nightly#_generate_examples
      signature = inspect.signature(self._split_generators)
      if "pipeline" in signature.parameters.keys():
        optional_pipeline_kwargs = dict(pipeline=split_builder.beam_pipeline)
      else:
        optional_pipeline_kwargs = {}
      split_generators = self._split_generators(  # pylint: disable=unexpected-keyword-arg
          dl_manager, **optional_pipeline_kwargs
      )
      # TODO(tfds): Could be removed once all datasets are migrated.
      # https://github.com/tensorflow/datasets/issues/2537
      # Legacy mode (eventually convert list[SplitGeneratorLegacy] -> dict)
      split_generators = split_builder.normalize_legacy_split_generators(
          split_generators=split_generators,
          generator_fn=self._generate_examples,
          is_beam=isinstance(self, BeamBasedBuilder),
      )

      # Ensure `all` isn't used as key.
      _check_split_names(split_generators.keys())

      # Writer fail if the number of example yield is `0`, so we return here.
      if download_config.max_examples_per_split == 0:
        return

      # Start generating data for all splits
      path_suffix = file_adapters.ADAPTER_FOR_FORMAT[
          self.info.file_format
      ].FILE_SUFFIX

      split_info_futures = []
      for split_name, generator in utils.tqdm(
          split_generators.items(),
          desc="Generating splits...",
          unit=" splits",
          leave=False,
      ):
        filename_template = naming.ShardedFileTemplate(
            split=split_name,
            dataset_name=self.name,
            data_dir=self.data_path,
            filetype_suffix=path_suffix,
        )
        future = split_builder.submit_split_generation(
            split_name=split_name,
            generator=generator,
            filename_template=filename_template,
            disable_shuffling=self.info.disable_shuffling,
        )
        split_info_futures.append(future)

    # Process the result of the beam pipeline.
    if maybe_pipeline_proxy._beam_pipeline:  # pylint:disable=protected-access
      self._process_pipeline_result(pipeline_result=maybe_pipeline_proxy.result)

    # Finalize the splits (after apache beam completed, if it was used)
    split_infos = [future.result() for future in split_info_futures]

    # Update the info object with the splits.
    split_dict = splits_lib.SplitDict(split_infos)
    self.info.set_splits(split_dict)

  def read_text_file(
      self, filename: epath.PathLike, encoding: Optional[str] = None
  ) -> str:
    """Returns the text in the given file and records the lineage."""
    filename = epath.Path(filename)
    self.info.add_file_data_source_access(filename)
    return filename.read_text(encoding=encoding)

  def read_tfrecord_as_dataset(
      self,
      filenames: str | Sequence[str],
      compression_type: str | None = None,
      num_parallel_reads: int | None = None,
  ) -> tf.data.Dataset:
    """Returns the dataset for the given tfrecord files and records the lineage."""
    if isinstance(filenames, str):
      filenames = [filenames]
    if not filenames:
      raise ValueError("No filenames given!")
    expanded_filenames = []
    for filename in filenames:
      self.info.add_file_data_source_access(filename)
      expanded_filenames.extend(file_utils.expand_glob(filename))
    filenames = [os.fspath(f) for f in expanded_filenames]

    return tf.data.TFRecordDataset(
        filenames=filenames,
        compression_type=compression_type,
        num_parallel_reads=num_parallel_reads,
    )

  def read_tfrecord_as_examples(
      self,
      filenames: Union[str, Sequence[str]],
      compression_type: str | None = None,
      num_parallel_reads: int | None = None,
  ) -> Iterator[tf.train.Example]:
    """Returns tf.Examples from the given tfrecord files and records the lineage."""
    raw_dataset = self.read_tfrecord_as_dataset(
        filenames=filenames,
        compression_type=compression_type,
        num_parallel_reads=num_parallel_reads,
    )
    for serialized_example in raw_dataset:
      example = tf.train.Example()
      example.ParseFromString(serialized_example.numpy())
      yield example

  def read_tfrecord_beam(
      self,
      file_pattern: str,
      /,
      **kwargs,
  ) -> "beam.PTransform":
    """Returns a PTransform reading the TFRecords and records it in the dataset lineage.

    This function records the lineage in the DatasetInfo and then invokes
    `beam.io.ReadFromTFRecord`. The kwargs should contain any other parameters
    for `beam.io.ReadFromTFRecord`. See
    https://beam.apache.org/releases/pydoc/2.6.0/apache_beam.io.tfrecordio.html#apache_beam.io.tfrecordio.ReadFromTFRecord.

    Arguments:
      file_pattern: A file glob pattern to read TFRecords from.
      **kwargs: the other parameters for `beam.io.ReadFromTFRecord`.

    Returns:
      a Beam PTransform that reads the given TFRecord files.
    """
    self.info.add_file_data_source_access(file_pattern)
    beam = lazy_imports_lib.lazy_imports.apache_beam
    return beam.io.ReadFromTFRecord(
        file_pattern=file_pattern,
        **kwargs,
    )


@utils.docs.deprecated
class BeamBasedBuilder(GeneratorBasedBuilder):
  """Beam based Builder.

  DEPRECATED: Please use `tfds.core.GeneratorBasedBuilder` instead.
  """

  def _generate_examples(
      self, *args: Any, **kwargs: Any
  ) -> split_builder_lib.SplitGenerator:
    return self._build_pcollection(*args, **kwargs)


def _check_split_names(split_names: Iterable[str]) -> None:
  """Check that split names are valid."""
  if "all" in set(str(s).lower() for s in split_names):
    raise ValueError(
        "`all` is a reserved keyword. Split cannot be named like this."
    )


def _get_default_config(
    builder_configs: List[BuilderConfig],
    default_config_name: Optional[str],
) -> Optional[BuilderConfig]:
  """Returns the default config from the given builder configs.

  Arguments:
    builder_configs: the configs from which the default should be picked.
    default_config_name: the name of the default config. If `None`, then it will
      use the first config.

  Returns:
    the default config. If there are no builder configs given, then None is
    returned.

  Raises:
    RuntimeError: if builder configs and a default config name are given, but no
    builder config with that name can be found.
  """
  if not builder_configs:
    return None
  if default_config_name is None:
    return builder_configs[0]
  for builder_config in builder_configs:
    if builder_config.name == default_config_name:
      return builder_config
  raise RuntimeError(
      f"Builder config with name `{default_config_name}` not found."
  )


def _save_default_config_name(
    common_dir: epath.Path,
    *,
    default_config_name: str,
) -> None:
  """Saves `builder_cls` metadata (common to all builder configs)."""
  data = {
      "default_config_name": default_config_name,
  }
  # `data_dir/ds_name/config/version/` -> `data_dir/ds_name/.config`
  config_dir = common_dir / ".config"
  config_dir.mkdir(parents=True, exist_ok=True)
  # Note:
  # * Save inside a dir to support some replicated filesystem
  # * Write inside a `.incomplete` file and rename to avoid multiple configs
  #   writing concurrently the same file
  # * Config file is overwritten each time a config is generated. If the
  #   default config is changed, this will be updated.
  config_path = config_dir / "metadata.json"
  with utils.incomplete_file(config_path) as tmp_config_path:
    tmp_config_path.write_text(json.dumps(data))


def load_default_config_name(
    common_dir: epath.Path,
) -> Optional[str]:
  """Load `builder_cls` metadata (common to all builder configs)."""
  config_path = epath.Path(common_dir) / ".config/metadata.json"
  if not config_path.exists():
    return None
  data = json.loads(config_path.read_text())
  return data.get("default_config_name")


def canonical_version_for_config(
    instance_or_cls: Union[DatasetBuilder, Type[DatasetBuilder]],
    config: Optional[BuilderConfig] = None,
) -> utils.Version:
  """Get the canonical version for the given config.

  This allow to get the version without instanciating the class.
  The version can be stored either at the class or in the config object.

  Args:
    instance_or_cls: The instance or class on which get the version
    config: The config which might contain the version, or None if the dataset
      do not have config.

  Returns:
    version: The extracted version.
  """
  if instance_or_cls.BUILDER_CONFIGS and config is None:
    raise ValueError(
        f"Cannot infer version on {instance_or_cls.name}. Unknown config."
    )

  if config and config.version:
    return utils.Version(config.version)
  elif instance_or_cls.VERSION:
    return utils.Version(instance_or_cls.VERSION)
  else:
    raise ValueError(
        f"DatasetBuilder {instance_or_cls.name} does not have a defined "
        "version. Please add a `VERSION = tfds.core.Version('x.y.z')` to the "
        "class."
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Waymo Open Dataset. See waymo.com/open."""

from __future__ import annotations

import os

from absl import logging
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.proto import waymo_dataset_pb2 as open_dataset
import tensorflow_datasets.public_api as tfds

_CITATION = """
@InProceedings{Sun_2020_CVPR,
author = {Sun, Pei and Kretzschmar, Henrik and Dotiwalla, Xerxes and Chouard, Aurelien and Patnaik, Vijaysai and Tsui, Paul and Guo, James and Zhou, Yin and Chai, Yuning and Caine, Benjamin and Vasudevan, Vijay and Han, Wei and Ngiam, Jiquan and Zhao, Hang and Timofeev, Aleksei and Ettinger, Scott and Krivokon, Maxim and Gao, Amy and Joshi, Aditya and Zhang, Yu and Shlens, Jonathon and Chen, Zhifeng and Anguelov, Dragomir},
title = {Scalability in Perception for Autonomous Driving: Waymo Open Dataset},
booktitle = {The IEEE/CVF Conference on Computer Vision and Pattern Recognition (CVPR)},
month = {June},
year = {2020}
}
"""

_DESCRIPTION = """\
The Waymo Open Dataset is comprised of high resolution sensor data
collected by Waymo self-driving cars in a wide variety of conditions.
This data is licensed for non-commercial use.

WARNING: this dataset requires additional authorization and registration.
Please look at tfds documentation for accessing GCS, and
afterwards, please register via https://waymo.com/open/licensing/
"""

_GCS_DESCRIPTION = """
This dataset is also available in pre-processed format, making it faster
to load, if you select the correct data_dir:

```
tfds.load('waymo_open_dataset/{}', \
data_dir='gs://waymo_open_dataset_{}_individual_files/tensorflow_datasets')
```

"""

_HOMEPAGE_URL = "http://www.waymo.com/open/"
_OBJECT_LABELS = [
    "TYPE_UNKNOWN",
    "TYPE_VEHICLE",
    "TYPE_PEDESTRIAN",
    "TYPE_SIGN",
    "TYPE_CYCLIST",
]


class WaymoOpenDatasetConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Waymo Open Dataset Config."""

  def __init__(
      self, *, name, version_str, description, is_on_gcs=False, **kwargs
  ):
    """BuilderConfig for Waymo Open Dataset examples.

    Args:
      name: Config name
      version_str: Version string (e.g. `v_1_2_0`).
      description: Description
      is_on_gcs: Whether the dataset is availabe preprocessed on GCS
      **kwargs: keyword arguments forwarded to super.
    """
    if is_on_gcs:
      description = description + _GCS_DESCRIPTION.format(name, version_str)
    super(WaymoOpenDatasetConfig, self).__init__(
        name=name,
        version=tfds.core.Version("0.2.0"),
        description=description,
        **kwargs,
    )
    self.cloud_bucket = tfds.core.Path(
        f"gs://waymo_open_dataset_{version_str}_individual_files/"
    )


class WaymoOpenDataset(tfds.core.BeamBasedBuilder):
  """Waymo Open Dataset."""

  BUILDER_CONFIGS = [
      WaymoOpenDatasetConfig(
          name="v1.2",
          version_str="v_1_2_0",
          description="Waymo Open Dataset v1.2",
      ),
      WaymoOpenDatasetConfig(
          name="v1.1",
          version_str="v_1_1_0",
          description="Waymo Open Dataset v1.1",
      ),
      WaymoOpenDatasetConfig(
          name="v1.0",
          version_str="v_1_0_0",
          description="Waymo Open Dataset v1.0",
          is_on_gcs=True,
      ),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    # Annotation descriptions are in the object development kit.
    annotations = {
        "type": tfds.features.ClassLabel(names=_OBJECT_LABELS),
        "bbox": tfds.features.BBoxFeature(),
    }

    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "context": {"name": tfds.features.Text()},
            "timestamp_micros": np.int64,
            "camera_FRONT": {
                "image": tfds.features.Image(
                    shape=(1280, 1920, 3), encoding_format="jpeg"
                ),
                "labels": tfds.features.Sequence(annotations),
            },
            "camera_FRONT_LEFT": {
                "image": tfds.features.Image(
                    shape=(1280, 1920, 3), encoding_format="jpeg"
                ),
                "labels": tfds.features.Sequence(annotations),
            },
            "camera_SIDE_LEFT": {
                "image": tfds.features.Image(
                    shape=(886, 1920, 3), encoding_format="jpeg"
                ),
                "labels": tfds.features.Sequence(annotations),
            },
            "camera_FRONT_RIGHT": {
                "image": tfds.features.Image(
                    shape=(1280, 1920, 3), encoding_format="jpeg"
                ),
                "labels": tfds.features.Sequence(annotations),
            },
            "camera_SIDE_RIGHT": {
                "image": tfds.features.Image(
                    shape=(886, 1920, 3), encoding_format="jpeg"
                ),
                "labels": tfds.features.Sequence(annotations),
            },
        }),
        homepage=_HOMEPAGE_URL,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns the SplitGenerators.

    Args:
      dl_manager: Download manager object.

    Returns:
      SplitGenerators.
    """

    # Training set
    train_files = tf.io.gfile.glob(
        os.path.join(
            self.builder_config.cloud_bucket, "training/segment*camera*"
        )
    )
    logging.info("Train files: %s", train_files)

    # Validation set
    validation_files = tf.io.gfile.glob(
        os.path.join(
            self.builder_config.cloud_bucket, "validation/segment*camera*"
        )
    )
    logging.info("Validation files: %s", validation_files)

    split_generators = [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "tf_record_files": train_files,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "tf_record_files": validation_files,
            },
        ),
    ]

    # Testing set (Only available in Waymo Open Dataset v1.2)
    if self.builder_config.name == "v_1_2":
      test_files = tf.io.gfile.glob(
          os.path.join(
              self.builder_config.cloud_bucket, "testing/segment*camera*"
          )
      )
      logging.info("Testing files: %s", test_files)

      split_generators.append(
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  "tf_record_files": test_files,
              },
          )
      )

    return split_generators

  def _build_pcollection(self, pipeline, tf_record_files):
    """Generate examples as dicts.

    Args:
      pipeline: Apache Beam pipeline.
      tf_record_files: .tfrecord files.

    Returns:
      Dict of examples.
    """
    beam = tfds.core.lazy_imports.apache_beam

    def _process_example(tf_record_file):
      for image_and_annotation in _generate_images_and_annotations(
          tf_record_file
      ):
        key = "%s:%s" % (
            image_and_annotation["context"]["name"],
            image_and_annotation["timestamp_micros"],
        )
        yield key, image_and_annotation

    return (
        pipeline | beam.Create(tf_record_files) | beam.FlatMap(_process_example)
    )


def _generate_images_and_annotations(tf_record_file):
  """Yields the images and annotations from a given file.

  Args:
    tf_record_file: .tfrecord files.

  Yields:
    Waymo images and annotations.
  """
  # Go through all frames
  dataset = tf.data.TFRecordDataset(tf_record_file, compression_type="")
  for data in dataset:
    frame = open_dataset.Frame()
    frame.ParseFromString(data.numpy())

    image_and_annotation = {
        "context": {"name": frame.context.name},
        "timestamp_micros": frame.timestamp_micros,
    }

    camera_calibration = {
        calibration.name: calibration
        for calibration in frame.context.camera_calibrations
    }
    camera_labels = {label.name: label for label in frame.camera_labels}

    # Go through all 5 camera images in the frame
    for frame_image in frame.images:
      labels = None
      if frame_image.name in camera_labels:
        image_height = camera_calibration[frame_image.name].height
        image_width = camera_calibration[frame_image.name].width
        labels = _convert_labels(
            camera_labels[frame_image.name], image_width, image_height
        )

      camera_name = open_dataset.CameraName.Name.Name(frame_image.name)
      image_and_annotation["camera_" + camera_name] = {
          "image": frame_image.image,
          "labels": labels,
      }

    yield image_and_annotation


def _convert_labels(raw_labels, image_width, image_height):
  """Convert labels to bounding boxes.

  Args:
    raw_labels: Raw label data.
    image_width: Width of the Waymo images.
    image_height: Height of the Waymo images.

  Returns:
    List of dicts with the label type and the corresponding bounding boxes.
  """
  return [
      {  # pylint: disable=g-complex-comprehension
          "type": raw_label.type,
          "bbox": _build_bounding_box(raw_label.box, image_width, image_height),
      }
      for raw_label in raw_labels.labels
  ]


def _build_bounding_box(open_dataset_box, image_width, image_height):
  """Builds and returns TFDS bounding box.

  Args:
    open_dataset_box: Bounding box center x,y coordinates and its length, width.
    image_width: Width of the Waymo images.
    image_height: Height of the Waymo images.

  Returns:
    tfds.features.BBox.
  """

  center_x = open_dataset_box.center_x
  center_y = open_dataset_box.center_y
  length = open_dataset_box.length
  width = open_dataset_box.width

  return tfds.features.BBox(
      ymin=max((center_y - (width / 2)) / image_height, 0.0),
      ymax=min((center_y + (width / 2)) / image_height, 1.0),
      xmin=max((center_x - (length / 2)) / image_width, 0.0),
      xmax=min((center_x + (length / 2)) / image_width, 1.0),
  )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base DatasetCollectionTestCase to test a DatasetCollection."""
import typing
from typing import List, Optional, Type

import pytest

from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_collection_builder
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered


class DatasetCollectionTestBase:
  """Inherit this class to test your DatasetCollection class.

  You must set the following class attribute:
    * DATASET_COLLECTION_CLASS: class object of DatasetCollection to test.

  You may set the following class attribute:
    * VERSION: The version of the DatasetCollection used to run the test.
      Defaults to None (latest version).
    * DATASETS_TO_TEST: List containing the datasets to test existence for. If
      no dataset is specified, the existence of all datasets in the collection
      will be tested.
    * CHECK_DATASETS_VERSION: Whether to check for the existence of the
      versioned datasets in the dataset collection, or for their default
      versions. Defaults to True.

  This test case will check for the following:
  - The dataset collection is correctly registered, i.e.
    `tfds.dataset_collection` works.
  - The datasets in the dataset collection exist, i.e. their builder is
    registered in tfds. If DATASETS_TO_TEST is not specified, all datasets in
    the collection will be checked.
  """

  DATASET_COLLECTION_CLASS = None
  VERSION: Optional[str] = None
  DATASETS_TO_TEST: List[str] = []
  CHECK_DATASETS_VERSION: bool = True

  @pytest.fixture(autouse=True)
  def dataset_collection(self):
    dataset_collection_cls = registered.imported_dataset_collection_cls(
        self.DATASET_COLLECTION_CLASS.name
    )
    dataset_collection_cls = typing.cast(
        Type[dataset_collection_builder.DatasetCollection],
        dataset_collection_cls,
    )
    yield dataset_collection_cls()

  @pytest.fixture(autouse=True)
  def datasets(self, dataset_collection):
    yield dataset_collection.get_collection(self.VERSION)

  def test_dataset_collection_is_registered(self, dataset_collection):
    """Checks that the dataset collection is registered."""
    assert registered.is_dataset_collection(dataset_collection.info.name)

  def test_dataset_collection_info(self, dataset_collection):
    """Checks that the collection's info is of type `DatasetCollectionInfo`."""
    assert isinstance(
        dataset_collection.info,
        dataset_collection_builder.DatasetCollectionInfo,
    )
    assert dataset_collection.info.name
    assert dataset_collection.info.description
    assert dataset_collection.info.release_notes

  def _get_dataset_builder(
      self, ds_reference: naming.DatasetReference, check_ds_version: bool = True
  ) -> dataset_builder.DatasetBuilder:
    """Returns the dataset builder for the requested dataset.

    Args:
      ds_reference: The `DatasetReference` containing the information about the
        dataset whose builder is to return.
      check_ds_version: Whether to return the dataset builder relative to a
        specific version of the dataset. If set to False, it will return the
        dataset builder for the latest version. Default is True.

    Returns:
      The dataset_builder.DatasetBuilder for the requested dataset.
    """
    cls = registered.imported_builder_cls(ds_reference.dataset_name)
    cls = typing.cast(Type[dataset_builder.DatasetBuilder], cls)
    version = ds_reference.version if check_ds_version else None
    return cls(config=ds_reference.config, version=version)

  def test_all_datasets_exist(self, datasets):
    """Checks that all the collection's datasets have a valid builder."""
    # If DATASETS_TO_TEST defined, restrict the scope of the datasets to test.
    if self.DATASETS_TO_TEST:
      datasets_to_test = {ds: datasets[ds] for ds in self.DATASETS_TO_TEST}
    else:
      datasets_to_test = datasets

    for ds, ds_reference in datasets_to_test.items():
      ds_builder = self._get_dataset_builder(
          ds_reference, check_ds_version=self.CHECK_DATASETS_VERSION
      )
      assert isinstance(ds_builder, dataset_builder.DatasetBuilder)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Kaggle API."""

import os

import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core.download import kaggle


class KaggleTest(testing.TestCase):

  def test_competition_download(self):
    with testing.mock_kaggle_api():
      with testing.tmp_dir() as tmp_dir:
        out_path = kaggle.download_kaggle_data('digit-recognizer', tmp_dir)
        self.assertEqual(
            os.fspath(out_path), os.path.join(tmp_dir, 'digit-recognizer')
        )
        with tf.io.gfile.GFile(os.path.join(out_path, 'output.txt')) as f:
          self.assertEqual('digit-recognizer', f.read())

  def test_dataset_download(self):
    with testing.mock_kaggle_api():
      with testing.tmp_dir() as tmp_dir:
        out_path = kaggle.download_kaggle_data('user/dataset', tmp_dir)
        self.assertIsInstance(out_path, os.PathLike)
        self.assertEqual(
            os.fspath(out_path), os.path.join(tmp_dir, 'user_dataset')
        )
        with tf.io.gfile.GFile(os.path.join(out_path, 'output.txt')) as f:
          self.assertEqual('user/dataset', f.read())

  def test_competition_download_404(self):
    with testing.mock_kaggle_api(err_msg='404 - Not found'):
      with testing.tmp_dir() as tmp_dir:
        with self.assertRaisesRegex(
            ValueError, 'Please ensure you have spelled the name'
        ):
          kaggle.download_kaggle_data('digit-recognize', tmp_dir)

  def test_kaggle_type(self):
    self.assertEqual(
        kaggle._get_kaggle_type('digit-recognizer'), 'competitions'
    )
    self.assertEqual(kaggle._get_kaggle_type('author/dataset'), 'datasets')


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Iterable

import apache_beam as beam
from apache_beam.testing.test_pipeline import TestPipeline
import numpy as np
import tensorflow as tf

from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core.dataset_builders import adhoc_builder
import tensorflow_datasets.public_api as tfds


class TfDataBuilderTest(testing.TestCase):

  def test_tf_data_builder_no_config(self):
    ds_name = 'configless_dataset'
    data_dir = self.tmp_dir

    builder = adhoc_builder.store_as_tfds_dataset(
        name=ds_name,
        version='1.0.0',
        split_datasets={
            'train': tf.data.Dataset.from_tensor_slices({
                'a': [1, 2],
                'b': ['a', 'b'],
            }),
            'test': tf.data.Dataset.from_tensor_slices({
                'a': [3],
                'b': ['c'],
            }),
        },
        features=tfds.features.FeaturesDict({
            'a': tf.int64,
            'b': tfds.features.Text(),
        }),
        data_dir=data_dir,
        disable_shuffling=True,
    )

    assert builder.data_dir == os.path.join(self.tmp_dir, ds_name, '1.0.0')
    assert set(builder.info.splits.keys()) == {'train', 'test'}
    assert builder.info.disable_shuffling

    actual_train = tfds.load(
        f'{ds_name}:1.0.0', split='train', data_dir=data_dir
    )
    actual_train_np = list(dataset_utils.as_numpy(actual_train))
    assert actual_train_np == [{'a': 1, 'b': b'a'}, {'a': 2, 'b': b'b'}]

    actual_test = tfds.load(f'{ds_name}:1.0.0', split='test', data_dir=data_dir)
    actual_test_np = list(dataset_utils.as_numpy(actual_test))
    assert actual_test_np == [{'a': 3, 'b': b'c'}]

    builder_from_dir = tfds.builder_from_directory(
        builder_dir=os.path.join(data_dir, ds_name, '1.0.0')
    )
    actual_train_from_dir = builder_from_dir.as_dataset(split='train')
    actual_train_from_dir_np = list(
        dataset_utils.as_numpy(actual_train_from_dir)
    )
    assert actual_train_from_dir_np == [
        {'a': 1, 'b': b'a'},
        {'a': 2, 'b': b'b'},
    ]

  def test_tf_data_builder_with_config(self):
    ds_name = 'my_dataset'
    data_dir = self.tmp_dir

    builder = adhoc_builder.store_as_tfds_dataset(
        name=ds_name,
        version='1.0.0',
        split_datasets={
            'train': tf.data.Dataset.from_tensor_slices({
                'a': [1, 2],
                'b': ['a', 'b'],
            }),
            'test': tf.data.Dataset.from_tensor_slices({
                'a': [3],
                'b': ['c'],
            }),
        },
        features=tfds.features.FeaturesDict({
            'a': tf.int64,
            'b': tfds.features.Text(),
        }),
        config=dataset_builder.BuilderConfig(name='some_config'),
        data_dir=data_dir,
    )

    assert builder.data_dir == os.path.join(
        self.tmp_dir, ds_name, 'some_config', '1.0.0'
    )
    assert set(builder.info.splits.keys()) == {'train', 'test'}

    actual_train = tfds.load(
        f'{ds_name}/some_config:1.0.0', split='train', data_dir=data_dir
    )
    actual_train_np = list(dataset_utils.as_numpy(actual_train))
    assert actual_train_np == [{'a': 1, 'b': b'a'}, {'a': 2, 'b': b'b'}]

    actual_test = tfds.load(
        f'{ds_name}/some_config:1.0.0', split='test', data_dir=data_dir
    )
    actual_test_np = list(dataset_utils.as_numpy(actual_test))
    assert actual_test_np == [{'a': 3, 'b': b'c'}]

    builder_from_dir = tfds.builder_from_directory(
        builder_dir=os.path.join(data_dir, ds_name, 'some_config', '1.0.0')
    )
    actual_train_from_dir = builder_from_dir.as_dataset(split='train')
    actual_train_from_dir_np = list(
        dataset_utils.as_numpy(actual_train_from_dir)
    )
    assert actual_train_from_dir_np == [
        {'a': 1, 'b': b'a'},
        {'a': 2, 'b': b'b'},
    ]

  def test_tf_data_builder_multiple_versions(self):
    ds_name = 'my_dataset'
    data_dir = self.tmp_dir
    builder_a = adhoc_builder.store_as_tfds_dataset(
        name=ds_name,
        version='1.0.0',
        split_datasets={
            'train': tf.data.Dataset.from_tensor_slices({'a': [1]})
        },
        features=tfds.features.FeaturesDict({'a': tf.int64}),
        data_dir=data_dir,
    )

    builder_b = adhoc_builder.store_as_tfds_dataset(
        name=ds_name,
        version='2.0.0',
        split_datasets={
            'train': tf.data.Dataset.from_tensor_slices({'a': [2]})
        },
        features=tfds.features.FeaturesDict({'a': tf.int64}),
        data_dir=data_dir,
    )

    assert builder_a.data_dir == os.path.join(self.tmp_dir, ds_name, '1.0.0')
    assert builder_b.data_dir == os.path.join(self.tmp_dir, ds_name, '2.0.0')

    train1 = tfds.load(f'{ds_name}:1.0.0', split='train', data_dir=data_dir)
    assert list(dataset_utils.as_numpy(train1)) == [{'a': 1}]

    train2 = tfds.load(f'{ds_name}:2.0.0', split='train', data_dir=data_dir)
    assert list(dataset_utils.as_numpy(train2)) == [{'a': 2}]

  def test_tf_data_builder_multiple_configs(self):
    ds_name = 'my_dataset'
    data_dir = self.tmp_dir
    builder_a = adhoc_builder.store_as_tfds_dataset(
        name=ds_name,
        version='1.0.0',
        config=dataset_builder.BuilderConfig(name='a'),
        split_datasets={
            'train': tf.data.Dataset.from_tensor_slices({'a': [1]})
        },
        features=tfds.features.FeaturesDict({'a': tf.int64}),
        data_dir=data_dir,
    )

    builder_b = adhoc_builder.store_as_tfds_dataset(
        name=ds_name,
        version='1.0.0',
        config=dataset_builder.BuilderConfig(name='b'),
        split_datasets={
            'train': tf.data.Dataset.from_tensor_slices({'a': [2]})
        },
        features=tfds.features.FeaturesDict({'a': tf.int64}),
        data_dir=data_dir,
    )

    assert builder_a.data_dir == os.path.join(
        self.tmp_dir, ds_name, 'a', '1.0.0'
    )
    assert builder_b.data_dir == os.path.join(
        self.tmp_dir, ds_name, 'b', '1.0.0'
    )

    train1 = tfds.load(f'{ds_name}/a:1.0.0', split='train', data_dir=data_dir)
    assert list(dataset_utils.as_numpy(train1)) == [{'a': 1}]

    train2 = tfds.load(f'{ds_name}/b:1.0.0', split='train', data_dir=data_dir)
    assert list(dataset_utils.as_numpy(train2)) == [{'a': 2}]

  def test_tf_data_builder_class(self):
    ds_name = 'configless_dataset'
    data_dir = self.tmp_dir

    builder = adhoc_builder.TfDataBuilder(
        name=ds_name,
        version='1.0.0',
        split_datasets={
            'train': tf.data.Dataset.from_tensor_slices({
                'a': [1, 2],
                'b': ['a', 'b'],
            }),
            'test': tf.data.Dataset.from_tensor_slices({
                'a': [3],
                'b': ['c'],
            }),
        },
        features=tfds.features.FeaturesDict({
            'a': tf.int64,
            'b': tfds.features.Text(),
        }),
        data_dir=data_dir,
    )
    builder.download_and_prepare()

    actual_train = tfds.load(
        f'{ds_name}:1.0.0', split='train', data_dir=data_dir
    )
    actual_train_np = list(dataset_utils.as_numpy(actual_train))
    assert actual_train_np == [{'a': 1, 'b': b'a'}, {'a': 2, 'b': b'b'}]

    actual_test = tfds.load(f'{ds_name}:1.0.0', split='test', data_dir=data_dir)
    actual_test_np = list(dataset_utils.as_numpy(actual_test))
    assert actual_test_np == [{'a': 3, 'b': b'c'}]


def generate_example(i: int, split: str):
  return i, {
      'id': i,
      'id_str': f'split={split} id={i}',
  }


def sort_ds_by_id(ds):
  return sorted(dataset_utils.as_numpy(ds), key=lambda x: x['id'])


class BeamBuilderTest(testing.TestCase):
  _DEFAULT_FEATURES = tfds.features.FeaturesDict({
      'id': tfds.features.Scalar(dtype=np.int64),
      'id_str': tfds.features.Text(),
  })

  def test_pcollection_is_not_iterable(self):
    with TestPipeline() as pipeline:
      d = pipeline | beam.Create(range(3))
      assert isinstance(d, beam.PCollection) or isinstance(
          d, beam.pvalue.PCollection
      )
      assert not isinstance(d, Iterable)

  def test_no_config(self):
    builder = adhoc_builder.store_as_tfds_dataset(
        name='adhoc',
        version='1.2.3',
        data_dir=self.tmp_dir,
        features=self._DEFAULT_FEATURES,
        split_datasets={
            'train': beam.Create(range(3)) | beam.Map(
                generate_example, split='train'
            ),
            'test': beam.Create(range(2)) | beam.Map(
                generate_example, split='test'
            ),
        },
    )

    assert builder.data_dir == os.path.join(self.tmp_dir, 'adhoc', '1.2.3')

    actual_train = tfds.load(
        'adhoc:1.2.3', split='train', data_dir=self.tmp_dir
    )
    assert sort_ds_by_id(actual_train) == [
        {'id': 0, 'id_str': b'split=train id=0'},
        {'id': 1, 'id_str': b'split=train id=1'},
        {'id': 2, 'id_str': b'split=train id=2'},
    ]

    actual_test = tfds.load('adhoc:1.2.3', split='test', data_dir=self.tmp_dir)
    assert sort_ds_by_id(actual_test) == [
        {'id': 0, 'id_str': b'split=test id=0'},
        {'id': 1, 'id_str': b'split=test id=1'},
    ]

  def test_config(self):
    builder_a = adhoc_builder.store_as_tfds_dataset(
        name='adhoc',
        version='1.2.3',
        config='a',
        data_dir=self.tmp_dir,
        features=self._DEFAULT_FEATURES,
        split_datasets={
            'train': beam.Create(range(3)) | beam.Map(
                generate_example, split='train'
            )
        },
    )

    builder_b = adhoc_builder.store_as_tfds_dataset(
        name='adhoc',
        version='1.2.3',
        config='b',
        data_dir=self.tmp_dir,
        features=self._DEFAULT_FEATURES,
        split_datasets={
            'train': beam.Create(range(2)) | beam.Map(
                generate_example, split='train'
            )
        },
    )

    assert builder_a.data_dir == os.path.join(
        self.tmp_dir, 'adhoc', 'a', '1.2.3'
    )
    assert builder_b.data_dir == os.path.join(
        self.tmp_dir, 'adhoc', 'b', '1.2.3'
    )

    actual_a = tfds.load('adhoc/a:1.2.3', split='train', data_dir=self.tmp_dir)
    assert sort_ds_by_id(actual_a) == [
        {'id': 0, 'id_str': b'split=train id=0'},
        {'id': 1, 'id_str': b'split=train id=1'},
        {'id': 2, 'id_str': b'split=train id=2'},
    ]

    actual_b = tfds.load('adhoc/b:1.2.3', split='train', data_dir=self.tmp_dir)
    assert sort_ds_by_id(actual_b) == [
        {'id': 0, 'id_str': b'split=train id=0'},
        {'id': 1, 'id_str': b'split=train id=1'},
    ]


def my_iterator(max_i: int, split: str):
  for i in range(max_i):
    yield generate_example(i, split)


class IteratorBuilderTest(testing.TestCase):
  _DEFAULT_FEATURES = tfds.features.FeaturesDict({
      'id': tfds.features.Scalar(dtype=np.int64),
      'id_str': tfds.features.Text(),
  })

  def test_no_config(self):
    builder = adhoc_builder.store_as_tfds_dataset(
        name='adhoc',
        version='1.2.3',
        data_dir=self.tmp_dir,
        features=self._DEFAULT_FEATURES,
        split_datasets={
            'train': my_iterator(max_i=3, split='train'),
            'test': my_iterator(max_i=2, split='test'),
        },
    )

    assert builder.data_dir == os.path.join(self.tmp_dir, 'adhoc', '1.2.3')
    assert set(builder.info.splits.keys()) == {'train', 'test'}

    actual_train = tfds.load(
        'adhoc:1.2.3', split='train', data_dir=self.tmp_dir
    )
    assert sort_ds_by_id(actual_train) == [
        {'id': 0, 'id_str': b'split=train id=0'},
        {'id': 1, 'id_str': b'split=train id=1'},
        {'id': 2, 'id_str': b'split=train id=2'},
    ]

    actual_test = tfds.load('adhoc:1.2.3', split='test', data_dir=self.tmp_dir)
    assert sort_ds_by_id(actual_test) == [
        {'id': 0, 'id_str': b'split=test id=0'},
        {'id': 1, 'id_str': b'split=test id=1'},
    ]

  def test_list(self):
    adhoc_builder.store_as_tfds_dataset(
        name='adhoc',
        version='1.2.3',
        data_dir=self.tmp_dir,
        features=self._DEFAULT_FEATURES,
        split_datasets={
            'train': list(my_iterator(max_i=1, split='train')),
        },
    )

    actual_train = tfds.load(
        'adhoc:1.2.3', split='train', data_dir=self.tmp_dir
    )
    assert sort_ds_by_id(actual_train) == [
        {'id': 0, 'id_str': b'split=train id=0'},
    ]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mozilla Common Voice Dataset."""

import csv
import dataclasses
import functools
import io
import re
from typing import Any, Mapping, Sequence

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DEMOGRAPHICS_URL = "https://github.com/common-voice/common-voice/blob/main/web/src/stores/demographics.ts"
_SPLITS = {
    tfds.Split.TRAIN: "train",
    tfds.Split.TEST: "test",
    tfds.Split.VALIDATION: "validated",
    tfds.Split("dev"): "dev",
}
_GENDER_CLASSES = ("male", "female", "other")

_DATA_FOLDER_IN_ARCHIVE = epath.Path("cv-corpus-6.1-2020-12-11")
_AUDIO_FILE_REGEX = re.compile(r"^.+/([^/]+\.mp3)$")

_LANGUAGES = (
    "en",  # Keep English as first to have it as the default config.
    "ab",
    "ar",
    "as",
    "br",
    "ca",
    "cnh",
    "cs",
    "cv",
    "cy",
    "de",
    "dv",
    "el",
    "eo",
    "es",
    "et",
    "eu",
    "fa",
    "fi",
    "fr",
    "fy-NL",
    "ga-IE",
    "hi",
    "hsb",
    "hu",
    "ia",
    "id",
    "it",
    "ja",
    "ka",
    "kab",
    "ky",
    "lg",
    "lt",
    "lv",
    "mn",
    "mt",
    "nl",
    "or",
    "pa-IN",
    "pl",
    "pt",
    "rm-sursilv",
    "rm-vallader",
    "ro",
    "ru",
    "rw",
    "sah",
    "sl",
    "sv-SE",
    "ta",
    "th",
    "tr",
    "tt",
    "uk",
    "vi",
    "vot",
    "zh-CN",
    "zh-HK",
    "zh-TW",
)

_BEAM_NAMESPACE = "TFDS_COMMON_VOICE"


def _download_url(language: str) -> str:
  return (
      "https://voice-prod-bundler-ee1969a6ce8178826482b88e843c335139bd3fb4."
      f"s3.amazonaws.com/cv-corpus-6.1-2020-12-11/{language}.tar.gz"
  )


class CommonVoiceConfig(tfds.core.BuilderConfig):
  """Configuration Class for Mozilla CommonVoice Dataset."""

  def __init__(self, *, language: str, **kwargs):
    """Constructs CommonVoiceConfig.

    Args:
     language: language Code of the Dataset to be used.
     **kwargs: keywords arguments forwarded to super
    """
    if language not in _LANGUAGES:
      raise ValueError(
          f"language {language} must be one of {', '.join(_LANGUAGES)}"
      )
    self.language = language

    super().__init__(
        name=language,
        description=f"Language Code: {language}",
        version=tfds.core.Version("2.0.0"),
        release_notes={
            "1.0.0": "Initial release.",
            "2.0.0": "Updated to corpus 6.1 from 2020-12-11.",
        },
        **kwargs,
    )


@dataclasses.dataclass()
class _IndexedExample:
  index: int
  clip_filename: str
  example: Mapping[str, Any]


def _parse_examples(examples_file_content: str) -> Sequence[_IndexedExample]:
  """Returns the examples in the given examples file content."""
  dataset = csv.DictReader(examples_file_content.splitlines(), delimiter="\t")
  examples = []
  for i, row in enumerate(dataset):
    path = row.get("path", None)
    if not path:
      continue
    try:
      example = {
          "client_id": row["client_id"],
          "sentence": row["sentence"],
          "upvotes": int(row["up_votes"]) if row["up_votes"] else 0,
          "downvotes": int(row["down_votes"]) if row["down_votes"] else 0,
          "age": row["age"],
          "gender": row["gender"] if row["gender"] else -1,
          "accent": row["accent"],
          "segment": row["segment"],
      }
      examples.append(
          _IndexedExample(index=i, example=example, clip_filename=path)
      )
    except KeyError as e:
      raise ValueError(f"Could not parse row: {row}") from e
  return examples


class CommonVoice(tfds.core.GeneratorBasedBuilder):
  """Mozilla Common Voice Dataset."""

  BUILDER_CONFIGS = tuple(CommonVoiceConfig(language=l) for l in _LANGUAGES)

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description="Mozilla Common Voice Dataset",
        features=tfds.features.FeaturesDict({
            "client_id": tfds.features.Text(doc="Hashed UUID of a given user"),
            "upvotes": tfds.features.Scalar(
                np.int32, doc="Number of people who said audio matches the text"
            ),
            "downvotes": tfds.features.Scalar(
                np.int32,
                doc="Number of people who said audio does not match text",
            ),
            "age": tfds.features.Text(
                doc=(
                    "Age bucket of the speaker (e.g. teens, or fourties), "
                    f"see {_DEMOGRAPHICS_URL}"
                )
            ),
            "gender": tfds.features.ClassLabel(
                names=_GENDER_CLASSES, doc="Gender of the speaker"
            ),
            "accent": tfds.features.Text(
                doc=f"Accent of the speaker, see {_DEMOGRAPHICS_URL}"
            ),
            "segment": tfds.features.Text(
                doc=(
                    "If sentence belongs to a custom dataset segment, "
                    "it will be listed here"
                )
            ),
            "sentence": tfds.features.Text(
                doc="Supposed transcription of the audio"
            ),
            "voice": tfds.features.Audio(),
        }),
        homepage="https://voice.mozilla.org/en/datasets",
        license=(
            "https://github.com/common-voice/common-voice/blob/main/LICENSE"
        ),
    )

  def _split_generators(self, dl_manager, pipeline):
    """Returns generators for each split in the dataset."""
    beam = tfds.core.lazy_imports.apache_beam

    language = self.builder_config.language
    dl_path = dl_manager.download(_download_url(language))

    examples_per_file = {}
    for path_within_archive, file_obj in dl_manager.iter_archive(dl_path):
      if not path_within_archive.endswith(".tsv"):
        continue
      examples_per_file[path_within_archive] = _parse_examples(
          file_obj.read().decode("utf-8")
      )

    clips_pipeline = self._clips_pipeline(
        pipeline=pipeline, name="clips", dl_manager=dl_manager, dl_path=dl_path
    )

    generators = {}
    generate_examples = beam.ptransform_fn(self._generate_examples)
    for split, split_filename in _SPLITS.items():
      filename = _DATA_FOLDER_IN_ARCHIVE / language / f"{split_filename}.tsv"
      examples = examples_per_file.get(str(filename), None)
      if not examples:
        continue
      generators[split] = pipeline | f"{split}-{language}" >> generate_examples(
          clips_pipeline=clips_pipeline, examples=examples
      )
    return generators

  def _clips_pipeline(self, pipeline, name, dl_manager, dl_path):
    beam = tfds.core.lazy_imports.apache_beam
    counter = functools.partial(beam.metrics.Metrics.counter, _BEAM_NAMESPACE)

    def _read_clips(dl_path):
      for filename, fobj in dl_manager.iter_archive(dl_path):
        if not filename.endswith(".mp3"):
          counter("clips_pipeline_non_mp3").inc()
          continue
        match = _AUDIO_FILE_REGEX.match(filename)
        if not match:
          counter("clips_pipeline_no_regex_match").inc()
          continue
        clip_filename = match.group(1)
        if not clip_filename:
          counter("clips_pipeline_empty_filename").inc()
          continue
        counter("clips_pipeline_clips").inc()
        yield clip_filename, io.BytesIO(fobj.read())

    return (
        pipeline
        | f"{name} - Collection of clips" >> beam.Create([dl_path])
        | f"{name} - Read clips" >> beam.FlatMap(_read_clips)
    )

  def _generate_examples(
      self,
      pipeline,
      clips_pipeline,
      examples: Sequence[_IndexedExample],
  ):
    """Generate Voice samples and statements given the data paths.

    Args:
      pipeline: Beam pipeline
      clips_pipeline: Beam pipeline with a stage that outputs a all audio clips
        keyed by the clip filename.
      examples: the examples in the split for which we need to generate
        examples.

    Returns:
      pipeline that produces dataset.
    """
    beam = tfds.core.lazy_imports.apache_beam
    counter = functools.partial(beam.metrics.Metrics.counter, _BEAM_NAMESPACE)

    def _process_join(el):
      try:
        (_, values) = el
        examples = values["examples"]
        if len(examples) != 1:
          counter(f"{len(examples)}_examples").inc()
          return
        indexed_example = examples[0]

        clips = values["clips"]
        if len(clips) != 1:
          counter("{len(clips)}_clips").inc()
          return
        clip = clips[0]

      except ValueError as e:
        raise ValueError(f"Unpacking element {el} failed") from e
      example = indexed_example.example.copy()
      example["voice"] = clip
      yield indexed_example.index, example

    examples_pipeline = pipeline | "Collection of examples" >> beam.Create(
        [[ex.clip_filename, ex] for ex in examples]
    )

    return (
        {"examples": examples_pipeline, "clips": clips_pipeline}
        | "Group by clip path" >> beam.CoGroupByKey()
        | "Reshuffle" >> beam.Reshuffle()
        | "Process and yield examples" >> beam.FlatMap(_process_join)
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Tool to generate the dataset community catalog documentation.
"""

import argparse
import dataclasses
import datetime
import functools
import json
import os
import re
import textwrap
from typing import Any, Iterator, List, Mapping, MutableMapping, Optional, Sequence, Tuple

from absl import app
from absl.flags import argparse_flags

from etils import epath
import tensorflow_datasets as tfds
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.community import register_package
from tensorflow_datasets.core.github_api import github_path
from tensorflow_datasets.core.proto import dataset_info_pb2
from tensorflow_datasets.core.proto import feature_pb2
from tensorflow_datasets.core.utils import gcs_utils
from tensorflow_datasets.core.utils import py_utils
import yaml

DatasetPackage = register_package.DatasetPackage

_INDEX_TEMPLATE_PATH = (
    'scripts/documentation/templates/community_catalog_overview.md'
)
_DATASET_DETAILS_TEMPLATE_PATH = (
    'scripts/documentation/templates/community_namespace_dataset.md'
)
_NAMESPACE_TOC_TEMPLATE_PATH = (
    'scripts/documentation/templates/community_namespace_toc.md'
)


@dataclasses.dataclass()
class _Options:
  catalog_dir: epath.Path
  local_cache: Optional[str] = None


def _parse_flags(_: List[str]) -> argparse.Namespace:
  """Command line flags."""
  parser = argparse_flags.ArgumentParser(
      prog='build_community_catalog',
      description='Tool to generate the community dataset catalog',
  )
  parser.add_argument(
      '--catalog_dir',
      help='Directory path where to generate the catalog. Default to TFDS dir.',
  )
  parser.add_argument(
      '--local_cache',
      help=(
          'If specified, where to store temporarily downloaded files so that '
          'they can be reused when running multiple times.'
      ),
  )
  return parser.parse_args()


def _load_template(template_path: str) -> str:
  return tfds.core.tfds_path(template_path).read_text()


def _to_toc_yaml(
    *,
    title: str,
    relative_path: str,
    page: str,
) -> Mapping[str, Any]:
  return {
      'title': title,
      'path': os.path.join(relative_path, page),
  }


def _clean_up_text(text: str) -> str:
  """Removes illegal characters from the given text."""
  return re.sub(r'```', '', text).strip()


def _get_huggingface_features(config: Mapping[str, Any]) -> feature_pb2.Feature:
  """Parses a huggingface config to a Feature proto."""
  return feature_pb2.Feature(
      json_feature=feature_pb2.JsonFeature(
          json=json.dumps(config['features'], indent=4)
      )
  )


def _get_cached_copy(file_path: epath.Path, max_age_days: int) -> Optional[str]:
  if file_path.exists():
    stats = os.stat(file_path)
    modified_time = datetime.datetime.fromtimestamp(stats.st_mtime)
    if modified_time > datetime.datetime.now() - datetime.timedelta(
        days=max_age_days
    ):
      return file_path.read_text()
  return None


@dataclasses.dataclass()
class DocumentationTemplates:
  index_template: str
  dataset_details_template: str
  namespace_toc_template: str

  @classmethod
  def load(cls) -> 'DocumentationTemplates':
    return cls(
        index_template=_load_template(_INDEX_TEMPLATE_PATH),
        dataset_details_template=_load_template(_DATASET_DETAILS_TEMPLATE_PATH),
        namespace_toc_template=_load_template(_NAMESPACE_TOC_TEMPLATE_PATH),
    )


@dataclasses.dataclass()
class DatasetDocumentation:
  """Functionality to generate documentation for a community dataset."""

  dataset: DatasetPackage
  templates: DocumentationTemplates
  options: _Options

  @property
  def tfds_id(self) -> str:
    """Returns the ID to use when loading this dataset in TFDS."""
    return str(self.dataset.name)

  @property
  def name(self) -> str:
    return self.dataset.name.name

  @property
  def namespace(self) -> str:
    return self.dataset.name.namespace

  def code_url(self, title: str = 'Code') -> str:
    return f'[{title}]({self.dataset.source.root_path})'

  def to_namespace_overview(self) -> str:
    """Returns the entry to be shown in the overview for a single namespace."""
    template = textwrap.dedent('*  [{dataset}]({namespace}/{dataset}.md)')
    return template.format(
        dataset=self.name,
        namespace=self.namespace,
    )

  def to_details_markdown(self) -> str:
    """ "Markdown to be shown on the details page for the namespace."""
    extra_links = self.format_extra_links(prefix='*   ', infix='\n')
    details = self.templates.dataset_details_template.format(
        name=self.name,
        description=self.documentation(),
        namespace=self.namespace,
        tfds_id=self.tfds_id,
        references_bulleted_list=extra_links,
    )
    if len(details) < 2 * 1024 * 1024:
      return details
    return self.templates.dataset_details_template.format(
        name=self.name,
        description=self.documentation(keep_short=True),
        namespace=self.namespace,
        tfds_id=self.tfds_id,
        references_bulleted_list=extra_links,
    )

  def dataset_info_per_config(
      self,
  ) -> Mapping[str, dataset_info_pb2.DatasetInfo]:
    return {}

  def documentation(self, keep_short: bool = False) -> str:
    """Returns detailed documentation for all configs of this dataset."""
    # TODO(weide): if e.g. the description contains markdown chars, then it
    # messes up the page. Try escaping backticks or using code blocks.
    # TODO(weide): how to format citation?
    header_template = '## {config_name}'
    template = textwrap.dedent("""
      Use the following command to load this dataset in TFDS:

      ```python
      ds = tfds.load('{tfds_id}')
      ```

      *   **Description**:

      ```
      {description}
      ```

      *   **License**: {license}
      *   **Version**: {version}
      *   **Splits**:
      {splits}
      *   **Features**:
      {features}
    """)

    #       *   **Citation**: `{citation}`

    def format_splits(split_infos: Sequence[dataset_info_pb2.SplitInfo]) -> str:
      splits_str = ('\n').join(
          sorted(
              [
                  f"`'{split.name}'` | {sum(split.shard_lengths)}"
                  for split in split_infos
              ]
          )
      )
      return textwrap.dedent(f"""
            Split  | Examples
            :----- | -------:
            {py_utils.indent(splits_str, '            ')}
            """)

    def format_feature(feature: feature_pb2.Feature) -> str:
      if feature.HasField('json_feature'):
        return textwrap.dedent(f"""
```json
{feature.json_feature.json}
```
                               """)
      if feature.HasField('features_dict'):
        descriptions = [
            f'    *   `{name}`'
            for name, _ in feature.features_dict.features.items()
        ]
        return '\n'.join(descriptions)
      return ''

    def format_template(
        config_name: str, info: dataset_info_pb2.DatasetInfo
    ) -> str:
      if config_name == 'default':
        tfds_id = self.tfds_id
      else:
        tfds_id = f'{self.tfds_id}/{config_name}'
      if keep_short:
        features = ''
      else:
        features = format_feature(info.features)
      content = template.format(
          description=_clean_up_text(info.description),
          tfds_id=tfds_id,
          license=info.redistribution_info.license or 'No known license',
          version=info.version,
          splits=format_splits(info.splits),
          features=features,
          citation=info.citation,
      )
      if config_name == 'default':
        return content
      return '{header}\n\n{content}'.format(
          header=header_template.format(config_name=config_name),
          content=content,
      )

    # homepage=default_dataset['homepage'],
    config_descriptions = [
        format_template(config_name, info)
        for config_name, info in self.dataset_info_per_config().items()
    ]

    return '\n\n'.join(config_descriptions)

  def extra_links(self) -> Sequence[str]:
    """List of extra links (formatted in Markdown) relevant for this dataset."""
    return [self.code_url()]

  def format_extra_links(self, prefix: str, infix: str) -> str:
    return infix.join([f'{prefix}{link}' for link in self.extra_links()])

  def to_toc_markdown(self) -> str:
    """Markdown to be shown in the TOC in the overview page."""
    extra_links = self.format_extra_links(prefix='', infix=' / ')
    return f'{self.name} ({extra_links})'


@dataclasses.dataclass()
class GithubDatasetDocumentation(DatasetDocumentation):
  """Specialized formatter for datasets whose dataset builder code is on Github."""

  @property
  def gh_path(self) -> github_path.GithubPath:
    return github_path.GithubPath(self.dataset.source.root_path)

  def code_url(self, title: str = 'Code') -> str:
    return f'[{title}]({self.gh_path.as_human_friendly_url()})'


@dataclasses.dataclass()
class HuggingfaceDatasetDocumentation(GithubDatasetDocumentation):
  """Specialized formatter for Huggingface."""

  def huggingface_link(self) -> str:
    return f'[Huggingface](https://huggingface.co/datasets/{self.name})'

  @functools.cached_property
  def dataset_infos(self) -> utils.JsonValue:
    """Retrieves the dataset info from either a cached copy or from Github."""

    content = None
    tmp_file_path = None
    if self.options.local_cache:
      tmp_ds_path = epath.Path(self.options.local_cache) / self.name
      tmp_ds_path.mkdir(exist_ok=True, parents=True)
      tmp_file_path = tmp_ds_path / 'dataset_infos.json'
      content = _get_cached_copy(file_path=tmp_file_path, max_age_days=14)

    if not content:
      try:
        dataset_info_path = self.gh_path / 'dataset_infos.json'
        print(f'Retrieving {dataset_info_path.as_raw_url()}')
        content = github_path.get_content(dataset_info_path.as_raw_url())
        if tmp_file_path:
          tmp_file_path.write_text(content.decode())
      except FileNotFoundError:
        return {}
    return json.loads(content)

  def extra_links(self) -> Sequence[str]:
    return [self.code_url(), self.huggingface_link()]

  def _parse_dataset_info_proto(
      self, config_name: str, config: Mapping[str, Any]
  ) -> dataset_info_pb2.DatasetInfo:
    """Parses a DatasetInfo proto from the given Json."""

    splits = []
    for name, details in config['splits'].items():
      splits.append(
          dataset_info_pb2.SplitInfo(
              name=name,
              num_shards=1,
              shard_lengths=[details['num_examples']],
              num_bytes=details['num_bytes'],
          )
      )

    if isinstance(config['version'], dict):
      version = config['version']['version_str']
    elif isinstance(config['version'], str):
      version = config['version']
    return dataset_info_pb2.DatasetInfo(
        name=config_name,
        module_name=config_name,
        description=config['description'],
        version=version,
        citation=config['citation'],
        redistribution_info=dataset_info_pb2.RedistributionInfo(
            license=config['license']
        ),
        splits=splits,
        features=_get_huggingface_features(config),
    )

  def dataset_info_per_config(
      self,
  ) -> Mapping[str, dataset_info_pb2.DatasetInfo]:
    if not isinstance(self.dataset_infos, dict):
      return {}
    return {
        config_name: self._parse_dataset_info_proto(config_name, config)
        for config_name, config in self.dataset_infos.items()
    }


@dataclasses.dataclass()
class NamespaceFormatter:
  """A documentation formatter for a namespace and all its datasets."""

  namespace: str
  datasets: Sequence[DatasetPackage]
  templates: DocumentationTemplates
  options: _Options

  @property
  def name(self) -> str:
    return self.namespace.capitalize()

  @property
  def overview_page(self) -> str:
    return f'{self.namespace}.md'

  def sections(self) -> Sequence[DatasetDocumentation]:
    return [
        DatasetDocumentation(
            dataset=ds, templates=self.templates, options=self.options
        )
        for ds in self.datasets
    ]

  def to_namespace_overview(self) -> str:
    """Markdown with dataset details for the namespace dataset overview."""
    sections = '\n'.join(
        [section.to_namespace_overview() for section in self.sections()]
    )
    template = textwrap.dedent("""
      # {self.name} datasets


      {sections}
      """)
    return template.format(sections=sections)

  def to_toc_markdown(self) -> str:
    datasets_for_toc = '\n'.join(
        [f'*   {section.to_toc_markdown()}' for section in self.sections()]
    )
    return self.templates.namespace_toc_template.format(
        name=self.name,
        namespace=self.namespace,
        overview_page=self.overview_page,
        num_datasets=len(self.datasets),
        datasets_for_toc=datasets_for_toc,
    )

  def to_toc_yaml(self, toc_relative_path: str):
    return _to_toc_yaml(
        title=self.name, relative_path=toc_relative_path, page=self.namespace
    )


@dataclasses.dataclass()
class HuggingfaceFormatter(NamespaceFormatter):
  """Specialized formatter for Huggingface."""

  def sections(self) -> Sequence[DatasetDocumentation]:
    return [
        HuggingfaceDatasetDocumentation(
            dataset=ds, templates=self.templates, options=self.options
        )
        for ds in self.datasets
    ]

  def to_namespace_overview(self) -> str:
    for section in self.sections():
      section.documentation()
    sections = '\n'.join(
        [section.to_namespace_overview() for section in self.sections()]
    )
    template = textwrap.dedent("""\
      # Huggingface datasets

      Huggingface has forked TFDS and provides a lot of text datasets. See
      [here](https://huggingface.co/docs/datasets/) for more documentation.
      Next you can find the list of all the datasets that can be used with TFDS.


      {sections}
      """)
    return template.format(sections=sections)


def formatter_for(
    namespace: str,
    datasets: Sequence[DatasetPackage],
    templates: DocumentationTemplates,
    options: _Options,
) -> NamespaceFormatter:
  """Returns the formatter for the given namespace and datasets."""
  if namespace == 'huggingface':
    return HuggingfaceFormatter(
        namespace=namespace,
        datasets=datasets,
        templates=templates,
        options=options,
    )
  return NamespaceFormatter(
      namespace=namespace,
      datasets=datasets,
      templates=templates,
      options=options,
  )


def build_overview(
    formatter_per_namespace: Mapping[str, NamespaceFormatter],
    templates: DocumentationTemplates,
) -> str:
  """Builds and saves the overview page."""
  toc_elements = [
      formatter.to_toc_markdown()
      for namespace, formatter in formatter_per_namespace.items()
  ]
  toc = '\n'.join(toc_elements)
  return templates.index_template.format(toc=toc)


def build_namespace_details(
    formatter_per_namespace: Mapping[str, NamespaceFormatter]
) -> Iterator[Tuple[str, str]]:
  """Generates all the detail pages for all namespaces and datasets."""
  for namespace, formatter in formatter_per_namespace.items():
    namespace_overview_markdown = formatter.to_namespace_overview()
    yield namespace, namespace_overview_markdown


def build_namespace_dataset_details(
    formatter: NamespaceFormatter,
) -> Iterator[Tuple[str, str]]:
  """Generates all the detail pages for all namespaces and datasets."""
  for section in formatter.sections():
    yield section.name, section.to_details_markdown()


def build_toc_yaml(
    formatter_per_namespace: Mapping[str, NamespaceFormatter]
) -> Mapping[str, Any]:
  """Returns a mapping representing the TOC (to be converted into Yaml)."""
  toc_relative_path = '/datasets/community_catalog'
  sections = [
      _to_toc_yaml(
          title='Overview', relative_path=toc_relative_path, page='overview'
      )
  ]
  for _, formatter in sorted(formatter_per_namespace.items()):
    sections.append(formatter.to_toc_yaml(toc_relative_path))
  return {'toc': sections}


def build_and_save_community_catalog(options: _Options) -> None:
  """Builds and saves the catalog of community datasets."""
  templates = DocumentationTemplates.load()
  formatter_per_namespace = _get_formatter_per_namespace(
      templates=templates, options=options
  )

  overview = build_overview(formatter_per_namespace, templates)
  options.catalog_dir.joinpath('overview.md').write_text(overview)

  # Write the `_toc.yaml` TF documentation navigation bar
  toc_yaml = build_toc_yaml(formatter_per_namespace)
  with options.catalog_dir.joinpath('_toc.yaml').open('w') as f:
    yaml.dump(toc_yaml, f, default_flow_style=False)

  for namespace, details in build_namespace_details(formatter_per_namespace):
    namespace_file = options.catalog_dir / f'{namespace}.md'
    namespace_file.write_text(details)

  for namespace, formatter in formatter_per_namespace.items():
    namespace_folder = options.catalog_dir / namespace
    namespace_folder.mkdir(exist_ok=True)
    for dataset_name, markdown in build_namespace_dataset_details(formatter):
      dataset_file = namespace_folder / f'{dataset_name}.md'
      dataset_file.write_text(markdown)


def _get_formatter_per_namespace(
    templates: DocumentationTemplates,
    options: _Options,
) -> Mapping[str, NamespaceFormatter]:
  """Returns the formatter per namespace."""
  formatters = {}
  for namespace, datasets in _get_datasets_per_namespace().items():
    formatters[namespace] = formatter_for(
        namespace=namespace,
        datasets=datasets,
        templates=templates,
        options=options,
    )
  return formatters


def _get_datasets_per_namespace() -> Mapping[str, Sequence[DatasetPackage]]:
  """Retrieves community datasets from GCS and groups them per namespace."""
  content = epath.Path(gcs_utils.GCS_COMMUNITY_INDEX_PATH).read_text()
  datasets_per_namespace: MutableMapping[str, List[DatasetPackage]] = {}
  for line in content.splitlines():
    dataset_package = DatasetPackage.from_json(json.loads(line))
    namespace = dataset_package.name.namespace
    if not namespace:
      raise ValueError(f'No namespace was specified for {dataset_package}')
    if namespace not in datasets_per_namespace:
      datasets_per_namespace[namespace] = [dataset_package]
    else:
      datasets_per_namespace[namespace].append(dataset_package)
  return datasets_per_namespace


def main(args: argparse.Namespace):
  catalog_dir = args.catalog_dir or os.path.join(
      tfds.core.utils.tfds_write_path(),
      'docs',
      'community_catalog',
  )

  options = _Options(
      catalog_dir=epath.Path(catalog_dir), local_cache=args.local_cache or None
  )
  build_and_save_community_catalog(options=options)


if __name__ == '__main__':
  app.run(main, flags_parser=_parse_flags)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.writer."""

import json
import os
from typing import Optional
from unittest import mock

from etils import epath
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import example_parser
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import writer as writer_lib
from tensorflow_datasets.core.utils import shard_utils
from tensorflow_datasets.core.writer import _ShardSpec


class GetShardSpecsTest(testing.TestCase):
  # Here we don't need to test all possible reading configs, as this is tested
  # by shard_utils.py.

  def test_1bucket_6shards(self):
    specs = writer_lib._get_shard_specs(
        num_examples=8,
        total_size=16,
        bucket_lengths=[8],
        filename_template=naming.ShardedFileTemplate(
            dataset_name='bar',
            split='train',
            data_dir='/',
            filetype_suffix='tfrecord',
        ),
        shard_config=shard_utils.ShardConfig(num_shards=6),
    )
    self.assertEqual(
        specs,
        [
            # Shard#, path, from_bucket, examples_number, reading instructions.
            _ShardSpec(
                0,
                '/bar-train.tfrecord-00000-of-00006',
                '/bar-train.tfrecord-00000-of-00006_index.json',
                1,
                [
                    shard_utils.FileInstruction(
                        filename='0', skip=0, take=1, examples_in_shard=8
                    ),
                ],
            ),
            _ShardSpec(
                1,
                '/bar-train.tfrecord-00001-of-00006',
                '/bar-train.tfrecord-00001-of-00006_index.json',
                2,
                [
                    shard_utils.FileInstruction(
                        filename='0', skip=1, take=2, examples_in_shard=8
                    ),
                ],
            ),
            _ShardSpec(
                2,
                '/bar-train.tfrecord-00002-of-00006',
                '/bar-train.tfrecord-00002-of-00006_index.json',
                1,
                [
                    shard_utils.FileInstruction(
                        filename='0', skip=3, take=1, examples_in_shard=8
                    ),
                ],
            ),
            _ShardSpec(
                3,
                '/bar-train.tfrecord-00003-of-00006',
                '/bar-train.tfrecord-00003-of-00006_index.json',
                1,
                [
                    shard_utils.FileInstruction(
                        filename='0', skip=4, take=1, examples_in_shard=8
                    ),
                ],
            ),
            _ShardSpec(
                4,
                '/bar-train.tfrecord-00004-of-00006',
                '/bar-train.tfrecord-00004-of-00006_index.json',
                2,
                [
                    shard_utils.FileInstruction(
                        filename='0', skip=5, take=2, examples_in_shard=8
                    ),
                ],
            ),
            _ShardSpec(
                5,
                '/bar-train.tfrecord-00005-of-00006',
                '/bar-train.tfrecord-00005-of-00006_index.json',
                1,
                [
                    shard_utils.FileInstruction(
                        filename='0', skip=7, take=1, examples_in_shard=8
                    ),
                ],
            ),
        ],
    )

  def test_4buckets_2shards(self):
    specs = writer_lib._get_shard_specs(
        num_examples=8,
        total_size=16,
        bucket_lengths=[2, 3, 0, 3],
        filename_template=naming.ShardedFileTemplate(
            dataset_name='bar',
            split='train',
            data_dir='/',
            filetype_suffix='tfrecord',
        ),
        shard_config=shard_utils.ShardConfig(num_shards=2),
    )
    self.assertEqual(
        specs,
        [
            # Shard#, path, examples_number, reading instructions.
            _ShardSpec(
                0,
                '/bar-train.tfrecord-00000-of-00002',
                '/bar-train.tfrecord-00000-of-00002_index.json',
                4,
                [
                    shard_utils.FileInstruction(
                        filename='0', skip=0, take=2, examples_in_shard=2
                    ),
                    shard_utils.FileInstruction(
                        filename='1', skip=0, take=2, examples_in_shard=3
                    ),
                ],
            ),
            _ShardSpec(
                1,
                '/bar-train.tfrecord-00001-of-00002',
                '/bar-train.tfrecord-00001-of-00002_index.json',
                4,
                [
                    shard_utils.FileInstruction(
                        filename='1', skip=2, take=-1, examples_in_shard=3
                    ),
                    shard_utils.FileInstruction(
                        filename='3', skip=0, take=-1, examples_in_shard=3
                    ),
                ],
            ),
        ],
    )


def _read_records(path, file_format=file_adapters.DEFAULT_FILE_FORMAT):
  """Returns (files_names, list_of_records_in_each_file).

  Args:
    path: path to tfrecord, omitting suffix.
    file_format: format of the record files.
  """
  # Ignore _index.json files.
  paths = sorted(tf.io.gfile.glob('%s-*-of-*' % path))
  paths = [p for p in paths if not p.endswith(writer_lib._INDEX_PATH_SUFFIX)]
  all_recs = []
  for fpath in paths:
    all_recs.append(
        list(
            dataset_utils.as_numpy(
                file_adapters.ADAPTER_FOR_FORMAT[file_format].make_tf_data(
                    fpath
                )
            )
        )
    )
  return [os.path.basename(p) for p in paths], all_recs


def _read_indices(path):
  """Returns (files_name, list of index in each file).

  Args:
    path: path to index, omitting suffix.
  """
  paths = sorted(tf.io.gfile.glob('%s-*-of-*_index.json' % path))
  all_indices = []
  for path in paths:
    json_str = epath.Path(path).read_text()
    # parse it back into a proto.
    shard_index = json.loads(json_str)
    all_indices.append(list(shard_index['index']))
  return [os.path.basename(p) for p in paths], all_indices


class WriterTest(testing.TestCase):
  EMPTY_SPLIT_ERROR = 'No examples were yielded.'
  TOO_SMALL_SPLIT_ERROR = 'num_examples (1) < number_of_shards (2)'

  NUM_SHARDS = 5
  RECORDS_TO_WRITE = [
      (1, b'a'),
      (2, b'b'),
      (3, b'c'),
      (4, b'd'),
      (5, b'e'),
      (6, b'f'),
      (7, b'g'),
      (8, b'hi'),
  ]
  RECORDS_WITH_HOLES = [
      (1, b'a'),
      (2, b'b'),
      (3, b'c'),
      (4, b'd'),
      (50000, b'e'),
      (600000, b'f'),
      (7000000, b'g'),
      (80000000, b'hi'),
  ]
  SHARDS_CONTENT = [
      [b'f', b'g'],
      [b'd'],
      [b'a', b'b'],
      [b'hi'],
      [b'e', b'c'],
  ]
  SHARDS_CONTENT_NO_SHUFFLING = [
      [b'a', b'b'],
      [b'c'],
      [b'd', b'e'],
      [b'f'],
      [b'g', b'hi'],
  ]

  def _write(
      self,
      to_write,
      salt: str = '',
      dataset_name: str = 'foo',
      split: str = 'train',
      disable_shuffling=False,
      file_format=file_adapters.DEFAULT_FILE_FORMAT,
      shard_config: Optional[shard_utils.ShardConfig] = None,
  ):
    filetype_suffix = file_adapters.ADAPTER_FOR_FORMAT[file_format].FILE_SUFFIX
    filename_template = naming.ShardedFileTemplate(
        dataset_name=dataset_name,
        split=split,
        filetype_suffix=filetype_suffix,
        data_dir=self.tmp_dir,
    )
    shard_config = shard_config or shard_utils.ShardConfig(
        num_shards=self.NUM_SHARDS
    )
    writer = writer_lib.Writer(
        serializer=testing.DummySerializer('dummy specs'),
        filename_template=filename_template,
        hash_salt=salt,
        disable_shuffling=disable_shuffling,
        file_format=file_format,
        shard_config=shard_config,
    )
    for key, record in to_write:
      writer.write(key, record)
    return writer.finalize()

  def test_write_tfrecord(self):
    """Stores records as tfrecord in a fixed number of shards with shuffling."""
    path = os.path.join(self.tmp_dir, 'foo-train.tfrecord')
    shards_length, total_size = self._write(to_write=self.RECORDS_TO_WRITE)
    self.assertEqual(self.NUM_SHARDS, len(shards_length))
    self.assertEqual(
        shards_length, [len(shard) for shard in self.SHARDS_CONTENT]
    )
    self.assertEqual(total_size, 9)
    written_files, all_recs = _read_records(path)
    written_index_files, all_indices = _read_indices(path)
    self.assertEqual(
        written_files,
        [
            f'foo-train.tfrecord-{i:05d}-of-{self.NUM_SHARDS:05d}'
            for i in range(self.NUM_SHARDS)
            if shards_length[i]
        ],
    )
    self.assertEqual(all_recs, self.SHARDS_CONTENT)
    self.assertEmpty(written_index_files)
    self.assertEmpty(all_indices)

  def test_write_tfrecord_sorted_by_key(self):
    """Stores records as tfrecord in a fixed number of shards without shuffling."""
    path = os.path.join(self.tmp_dir, 'foo-train.tfrecord')
    shards_length, total_size = self._write(
        to_write=self.RECORDS_TO_WRITE, disable_shuffling=True
    )
    self.assertEqual(
        shards_length,
        [len(shard) for shard in self.SHARDS_CONTENT_NO_SHUFFLING],
    )
    self.assertEqual(total_size, 9)
    written_files, all_recs = _read_records(path)
    written_index_files, all_indices = _read_indices(path)
    self.assertEqual(
        written_files,
        [
            f'foo-train.tfrecord-{i:05d}-of-{self.NUM_SHARDS:05d}'
            for i in range(self.NUM_SHARDS)
            if shards_length[i]
        ],
    )
    self.assertEqual(all_recs, self.SHARDS_CONTENT_NO_SHUFFLING)
    self.assertEmpty(written_index_files)
    self.assertEmpty(all_indices)

  def test_write_tfrecord_sorted_by_key_with_holes(self):
    """Stores records as tfrecord in a fixed number of shards without shuffling."""
    path = os.path.join(self.tmp_dir, 'foo-train.tfrecord')
    shards_length, total_size = self._write(
        to_write=self.RECORDS_WITH_HOLES, disable_shuffling=True
    )
    self.assertEqual(
        shards_length,
        [len(shard) for shard in self.SHARDS_CONTENT_NO_SHUFFLING],
    )
    self.assertEqual(total_size, 9)
    written_files, all_recs = _read_records(path)
    written_index_files, all_indices = _read_indices(path)
    self.assertEqual(
        written_files,
        [
            f'foo-train.tfrecord-{i:05d}-of-{self.NUM_SHARDS:05d}'
            for i in range(self.NUM_SHARDS)
            if shards_length[i]
        ],
    )
    self.assertEqual(all_recs, self.SHARDS_CONTENT_NO_SHUFFLING)
    self.assertEmpty(written_index_files)
    self.assertEmpty(all_indices)

  @mock.patch.object(example_parser, 'ExampleParser', testing.DummyParser)
  def test_write_duplicated_keys(self):
    to_write = [(1, b'a'), (2, b'b'), (1, b'c')]
    with self.assertRaisesWithPredicateMatch(
        AssertionError, 'Two examples share the same hashed key'
    ):
      shard_config = shard_utils.ShardConfig(num_shards=1)
      self._write(to_write=to_write, shard_config=shard_config)

  def test_empty_split(self):
    to_write = []
    with self.assertRaisesWithPredicateMatch(
        AssertionError, self.EMPTY_SPLIT_ERROR
    ):
      shard_config = shard_utils.ShardConfig(num_shards=1)
      self._write(to_write=to_write, shard_config=shard_config)

  def test_too_small_split(self):
    to_write = [(1, b'a')]
    with self.assertRaisesWithPredicateMatch(
        AssertionError, self.TOO_SMALL_SPLIT_ERROR
    ):
      shard_config = shard_utils.ShardConfig(num_shards=2)
      self._write(to_write=to_write, shard_config=shard_config)
      self._write(to_write=to_write)


class TfrecordsWriterBeamTest(testing.TestCase):
  NUM_SHARDS = 3
  RECORDS_TO_WRITE = [(i, str(i).encode('utf-8')) for i in range(10)]
  SHARDS_CONTENT = [
      [b'6', b'9'],
      [b'7'],
      [b'4', b'1', b'2', b'8', b'0', b'5', b'3'],
  ]
  SHARDS_CONTENT_NO_SHUFFLING = [
      [b'0', b'1', b'2'],
      [b'3', b'4', b'5'],
      [b'6', b'7', b'8', b'9'],
  ]

  def _write(
      self,
      to_write,
      salt: str = '',
      dataset_name: str = 'foo',
      split: str = 'train',
      disable_shuffling=False,
      file_format=file_adapters.DEFAULT_FILE_FORMAT,
      shard_config: Optional[shard_utils.ShardConfig] = None,
  ):
    filetype_suffix = file_adapters.ADAPTER_FOR_FORMAT[file_format].FILE_SUFFIX
    filename_template = naming.ShardedFileTemplate(
        dataset_name=dataset_name,
        split=split,
        filetype_suffix=filetype_suffix,
        data_dir=self.tmp_dir,
    )
    shard_config = shard_config or shard_utils.ShardConfig(
        num_shards=self.NUM_SHARDS
    )
    beam = lazy_imports_lib.lazy_imports.apache_beam
    writer = writer_lib.BeamWriter(
        serializer=testing.DummySerializer('dummy specs'),
        filename_template=filename_template,
        hash_salt=salt,
        disable_shuffling=disable_shuffling,
        file_format=file_format,
        shard_config=shard_config,
    )
    # Here we need to disable type check as `beam.Create` is not capable of
    # inferring the type of the PCollection elements.
    options = beam.options.pipeline_options.PipelineOptions(
        pipeline_type_check=False
    )
    with beam.Pipeline(options=options) as pipeline:

      @beam.ptransform_fn
      def _build_pcollection(pipeline):
        pcollection = pipeline | 'Start' >> beam.Create(to_write)
        return writer.write_from_pcollection(pcollection)

      _ = pipeline | 'test' >> _build_pcollection()  # pylint: disable=no-value-for-parameter
    return writer.finalize()

  def test_write_tfrecord(self):
    """Stores records as tfrecord in a fixed number of shards with shuffling."""
    path = os.path.join(self.tmp_dir, 'foo-train.tfrecord')
    shards_length, total_size = self._write(to_write=self.RECORDS_TO_WRITE)
    self.assertEqual(self.NUM_SHARDS, len(shards_length))
    self.assertEqual(
        shards_length, [len(shard) for shard in self.SHARDS_CONTENT]
    )
    self.assertEqual(total_size, 10)
    written_files, all_recs = _read_records(path)
    written_index_files, all_indices = _read_indices(path)
    self.assertEqual(
        written_files,
        [
            f'foo-train.tfrecord-{i:05d}-of-{self.NUM_SHARDS:05d}'
            for i in range(self.NUM_SHARDS)
            if shards_length[i]
        ],
    )
    self.assertEqual(all_recs, self.SHARDS_CONTENT)
    self.assertEmpty(written_index_files)
    self.assertEmpty(all_indices)

  def test_empty_split(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError,
        'The total number of generated examples is 0. This should be >0!',
    ):
      self._write(to_write=[])

  def test_write_tfrecord_sorted_by_key(self):
    """Stores records as tfrecord in a fixed number of shards without shuffling."""
    path = os.path.join(self.tmp_dir, 'foo-train.tfrecord')
    shards_length, total_size = self._write(
        to_write=self.RECORDS_TO_WRITE, disable_shuffling=True
    )
    self.assertEqual(
        shards_length,
        [len(shard) for shard in self.SHARDS_CONTENT_NO_SHUFFLING],
    )
    self.assertEqual(total_size, 10)
    written_files, all_recs = _read_records(path)
    written_index_files, all_indices = _read_indices(path)
    self.assertEqual(
        written_files,
        [
            f'foo-train.tfrecord-{i:05d}-of-{self.NUM_SHARDS:05d}'
            for i in range(self.NUM_SHARDS)
            if shards_length[i]
        ],
    )
    self.assertEqual(all_recs, self.SHARDS_CONTENT_NO_SHUFFLING)
    self.assertEmpty(written_index_files)
    self.assertEmpty(all_indices)

  def test_write_tfrecord_sorted_by_key_with_holes(self):
    """Stores records as tfrecord in a fixed number of shards without shuffling.

    Note that the keys are not consecutive but contain gaps.
    """
    path = os.path.join(self.tmp_dir, 'foo-train.tfrecord')
    records_with_holes = [
        (i**4, str(i**4).encode('utf-8')) for i in range(10)
    ]
    expected_shards = [
        [b'0', b'1', b'16', b'81', b'256', b'625', b'1296'],
        [b'2401', b'4096'],
        [b'6561'],
    ]

    shards_length, total_size = self._write(
        to_write=records_with_holes, disable_shuffling=True
    )
    self.assertEqual(shards_length, [len(shard) for shard in expected_shards])
    self.assertEqual(total_size, 28)
    written_files, all_recs = _read_records(path)
    written_index_files, all_indices = _read_indices(path)
    self.assertEqual(
        written_files,
        [
            f'foo-train.tfrecord-{i:05d}-of-{self.NUM_SHARDS:05d}'
            for i in range(self.NUM_SHARDS)
        ],
    )
    self.assertEqual(all_recs, expected_shards)
    self.assertEmpty(written_index_files)
    self.assertEmpty(all_indices)


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""i_naturalist2018 dataset."""

from tensorflow_datasets.image_classification.i_naturalist2018 import i_naturalist2018
import tensorflow_datasets.public_api as tfds


class INaturalist2018Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for i_naturalist2018 dataset."""

  DATASET_CLASS = i_naturalist2018.INaturalist2018
  SPLITS = {
      "train": 4,  # Number of fake train example
      "validation": 3,  # Number of fake val example
      "test": 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      "test_images": "test2018.tar.gz",
      "train_annos": "train_annos",
      "val_annos": "val_annos",
      "trainval_images": "train_val2018.tar.gz",
      "categories": "categories",
  }
  SKIP_CHECKSUMS = True


if __name__ == "__main__":
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from tensorflow_datasets.core.utils import bool_utils


@pytest.mark.parametrize(
    ['string', 'result'],
    [
        (True, True),
        (False, False),
        # In `wit_kaggle` dataset
        ('true', True),
        (' true  ', True),
        ('false', False),
        ('  false ', False),
        # In `gap` dataset
        ('TRUE', True),
        ('FALSE', False),
        # In `kddcup99`/`duke_ultrasound_dataset_builder` datasets
        (0, False),
        ('0', False),
        (1, True),
        ('1', True),
    ],
)
def test_parse_bool_success(string, result):
  assert bool_utils.parse_bool(string) == result


@pytest.mark.parametrize(
    ['string', 'exception'],
    [
        ('_true', 'Cannot convert "_true" to bool'),
        ('2', 'Cannot convert "2" to bool'),
    ],
)
def test_parse_bool_exception(string, exception):
  with pytest.raises(Exception, match=exception):
    bool_utils.parse_bool(string)
